"""
Flight controller interface.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import os
from argparse import ArgumentParser
from logging import debug as logging_debug
from logging import error as logging_error
from logging import info as logging_info
from logging import warning as logging_warning
from os import name as os_name
from os import path as os_path
from os import readlink as os_readlink
from pathlib import Path
from time import sleep as time_sleep
from time import time as time_time
from typing import Callable, NoReturn, Optional, Union

import serial.tools.list_ports
import serial.tools.list_ports_common
from pymavlink import mavutil
from pymavlink.dialects.v20.ardupilotmega import MAVLink_autopilot_version_message
from serial.serialutil import SerialException

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.argparse_check_range import CheckRange
from ardupilot_methodic_configurator.backend_flightcontroller_info import BackendFlightcontrollerInfo
from ardupilot_methodic_configurator.backend_mavftp import MAVFTP
from ardupilot_methodic_configurator.data_model_par_dict import ParDict

# pylint: disable=too-many-lines


class FakeSerialForTests:
    """
    A mock serial class for unit testing purposes.

    This class simulates the behavior of a serial connection for testing purposes,
    allowing for the testing of serial communication without needing a physical
    serial device. It includes methods for reading, writing, and checking the
    number of bytes in the input buffer, as well as closing the connection.
    """

    def __init__(self, device: str) -> None:
        self.device = device

    def read(self, _len) -> str:  # noqa: ANN001
        return ""

    def write(self, _buf) -> NoReturn:  # noqa: ANN001
        msg = "write always fails"
        raise Exception(msg)  # pylint: disable=broad-exception-raised

    def inWaiting(self) -> int:  # noqa: N802, pylint: disable=invalid-name
        return 0

    def close(self) -> None:
        pass


DEFAULT_BAUDRATE: int = 115200
DEFAULT_REBOOT_TIME: int = 7
# https://github.com/ArduPilot/ardupilot/blob/master/libraries/AP_SerialManager/AP_SerialManager.cpp#L741C1-L757C32
SUPPORTED_BAUDRATES: list[str] = [
    "1200",
    "2400",
    "4800",
    "9600",
    "19200",
    "38400",
    "57600",
    "100000",
    "111100",
    "115200",
    "230400",
    "256000",
    "460800",
    "500000",
    "921600",
    "1500000",
    "2000000",
]


class FlightController:  # pylint: disable=too-many-public-methods,too-many-instance-attributes
    """
    A class to manage the connection and parameters of a flight controller.

    Attributes:
        device (str): The connection string to the flight controller.
        master (mavutil.mavlink_connection): The MAVLink connection object.
        fc_parameters (dict[str, float]): A dictionary of flight controller parameters.

    """

    def __init__(self, reboot_time: int = DEFAULT_REBOOT_TIME, baudrate: int = DEFAULT_BAUDRATE) -> None:
        """Initialize the FlightController communication object."""
        # warn people about ModemManager which interferes badly with ArduPilot
        if os_path.exists("/usr/sbin/ModemManager"):
            logging_warning(_("You should uninstall ModemManager as it conflicts with ArduPilot"))

        self.__reboot_time = reboot_time
        self.__baudrate = baudrate
        self.__connection_tuples: list[tuple[str, str]] = []
        self.discover_connections()
        self.master: Union[mavutil.mavlink_connection, None] = None  # pyright: ignore[reportGeneralTypeIssues]
        self.comport: Union[mavutil.SerialPort, None] = None
        self.fc_parameters: dict[str, float] = {}
        self.info = BackendFlightcontrollerInfo()

        # Battery status tracking
        self._last_battery_message_time: float = 0.0
        self._last_battery_status: Union[tuple[float, float], None] = None

    def discover_connections(self) -> None:
        comports = FlightController.__list_serial_ports()
        netports = FlightController.__list_network_ports()
        # list of tuples with the first element being the port name and the second element being the port description
        self.__connection_tuples = [(port.device, port.description) for port in comports] + [(port, port) for port in netports]
        logging_info(_("Available connection ports are:"))
        for port in self.__connection_tuples:
            logging_info("%s - %s", port[0], port[1])
        # now that it is logged, add the 'Add another' tuple
        self.__connection_tuples += [(_("Add another"), _("Add another"))]

    def disconnect(self) -> None:
        """Close the connection to the flight controller."""
        if self.master is not None:
            self.master.close()
            self.master = None
        self.fc_parameters = {}
        self.info = BackendFlightcontrollerInfo()

    def add_connection(self, connection_string: str) -> bool:
        """Add a new connection to the list of available connections."""
        if connection_string:
            # Check if connection_string is not the first element of any tuple in self.other_connection_tuples
            if all(connection_string != t[0] for t in self.__connection_tuples):
                self.__connection_tuples.insert(-1, (connection_string, connection_string))
                logging_debug(_("Added connection %s"), connection_string)
                return True
            logging_debug(_("Did not add duplicated connection %s"), connection_string)
        else:
            logging_debug(_("Did not add empty connection"))
        return False

    def _register_and_try_connect(
        self,
        comport: mavutil.SerialPort,
        progress_callback: Union[None, Callable[[int, int], None]],
        baudrate: int,
        log_errors: bool,
    ) -> str:
        """
        Register a device in the connection list (if missing) and attempt connection.

        Returns:
            str: empty string on success, or error message.

        """
        # set comport for subsequent calls
        self.comport = comport
        # Add the detected port to the list of available connections if it is not there
        if self.comport and self.comport.device not in [t[0] for t in self.__connection_tuples]:
            self.__connection_tuples.insert(-1, (self.comport.device, getattr(self.comport, "description", "")))
        # Try to connect
        return self.__create_connection_with_retry(
            progress_callback=progress_callback, baudrate=baudrate, log_errors=log_errors, timeout=2
        )

    def connect(
        self,
        device: str,
        progress_callback: Union[None, Callable[[int, int], None]] = None,
        log_errors: bool = True,
        baudrate: Optional[int] = None,
    ) -> str:
        """
        Establishes a connection to the FlightController using a specified device.

        This method attempts to connect to the FlightController using the provided device
        connection string. If no device is specified, it attempts to auto-detect a serial
        port that matches the preferred ports list. If no serial device is found it tries
        the "standard" ArduPilot UDP and TCP connections. If a device is specified as 'test',
        it sets some test parameters for debugging purposes.

        Args:
            device (str): The connection string to the flight controller. If an empty string
                        is provided, the method attempts to auto-detect a serial port.
            progress_callback (callable, optional): A callback function to report the progress
                                                    of the connection attempt. Default is None.
            log_errors: log errors
            baudrate (int, optional): The baudrate to use for the connection. If None,
                                    uses the default baudrate from initialization.

        Returns:
            str: An error message if the connection fails, otherwise an empty string indicating
                a successful connection.

        """
        connection_baudrate = baudrate
        if connection_baudrate is None:
            connection_baudrate = self.__baudrate

        if device:
            if device == "none":
                return ""
            self.add_connection(device)
            self.comport = mavutil.SerialPort(device=device, description=device)
            return self.__create_connection_with_retry(
                progress_callback=progress_callback, baudrate=connection_baudrate, log_errors=log_errors
            )

        # Try to autodetect serial ports
        autodetect_serial = self.__auto_detect_serial()
        if autodetect_serial:
            # Resolve the soft link if it's a Linux system
            if os_name == "posix":
                try:
                    dev = autodetect_serial[0].device
                    logging_debug(_("Auto-detected device %s"), dev)
                    # Get the directory part of the soft link
                    softlink_dir = os_path.dirname(dev)
                    # Resolve the soft link and join it with the directory part
                    resolved_path = os_path.abspath(os_path.join(softlink_dir, os_readlink(dev)))
                    autodetect_serial[0].device = resolved_path
                    logging_debug(_("Resolved soft link %s to %s"), dev, resolved_path)
                except OSError:
                    pass  # Not a soft link, proceed with the original device path
            err = self._register_and_try_connect(
                comport=autodetect_serial[-1],
                progress_callback=progress_callback,
                baudrate=connection_baudrate,
                log_errors=False,
            )
            if err == "":
                return ""

        # Try to autodetect network ports
        netports = FlightController.__list_network_ports()
        for port in netports:
            # try to connect to each "standard" ArduPilot UDP and TCP ports
            logging_debug(_("Trying network port %s"), port)
            err = self._register_and_try_connect(
                comport=mavutil.SerialPort(device=port, description=port),
                progress_callback=progress_callback,
                baudrate=self.__baudrate,
                log_errors=False,
            )
            if err == "":
                return ""

        return _("No auto-detected ports responded. Please connect a flight controller and try again.")

    def __request_banner(self) -> None:
        """Request banner information from the flight controller."""
        # https://mavlink.io/en/messages/ardupilotmega.html#MAV_CMD_DO_SEND_BANNER
        if self.master is not None:
            # Note: Don't wait for ACK here as banner requests are fire-and-forget
            # and we handle the response via STATUS_TEXT messages
            self.master.mav.command_long_send(
                self.master.target_system,
                self.master.target_component,
                mavutil.mavlink.MAV_CMD_DO_SEND_BANNER,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            )

    def __receive_banner_text(self) -> list[str]:
        """Starts listening for STATUS_TEXT MAVLink messages."""
        start_time = time_time()
        banner_msgs: list[str] = []
        while self.master:
            msg = self.master.recv_match(type="STATUSTEXT", blocking=False)
            if msg:
                if banner_msgs:
                    banner_msgs.append(msg.text)
                else:
                    banner_msgs = [msg.text]
            time_sleep(0.1)  # Sleep briefly to reduce CPU usage
            if time_time() - start_time > 1:  # Check if 1 seconds have passed since the start of the loop
                break  # Exit the loop if 1 seconds have elapsed
        return banner_msgs

    def __request_message(self, message_id: int) -> None:
        """Request a specific message from the flight controller."""
        if self.master is not None:
            # Note: Don't wait for ACK here as this is used internally for autopilot version requests
            # and the response comes as the requested message itself
            self.master.mav.command_long_send(
                self.info.system_id,
                self.info.component_id,
                mavutil.mavlink.MAV_CMD_REQUEST_MESSAGE,
                0,  # confirmation
                message_id,
                0,
                0,
                0,
                0,
                0,
                0,
            )

    def _send_command_and_wait_ack(  # pylint: disable=too-many-arguments,too-many-positional-arguments, too-many-locals
        self,
        command: int,
        param1: float = 0,
        param2: float = 0,
        param3: float = 0,
        param4: float = 0,
        param5: float = 0,
        param6: float = 0,
        param7: float = 0,
        timeout: float = 5.0,
    ) -> tuple[bool, str]:
        """
        Send a MAVLink command and wait for acknowledgment.

        Args:
            command: The MAVLink command ID
            param1: Command parameter 1
            param2: Command parameter 2
            param3: Command parameter 3
            param4: Command parameter 4
            param5: Command parameter 5
            param6: Command parameter 6
            param7: Command parameter 7
            timeout: Timeout in seconds to wait for acknowledgment

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for command")
            logging_error(error_msg)
            return False, error_msg

        try:
            # Send the command
            self.master.mav.command_long_send(
                self.master.target_system,
                self.master.target_component,
                command,
                0,  # confirmation
                param1,
                param2,
                param3,
                param4,
                param5,
                param6,
                param7,
            )

            # Wait for acknowledgment
            start_time = time_time()
            while time_time() - start_time < timeout:
                msg = self.master.recv_match(type="COMMAND_ACK", blocking=False)
                if msg and msg.command == command:
                    # Map result codes to error messages
                    result_messages = {
                        mavutil.mavlink.MAV_RESULT_ACCEPTED: ("", True),
                        mavutil.mavlink.MAV_RESULT_TEMPORARILY_REJECTED: (_("Command temporarily rejected"), False),
                        mavutil.mavlink.MAV_RESULT_DENIED: (_("Command denied"), False),
                        mavutil.mavlink.MAV_RESULT_UNSUPPORTED: (_("Command unsupported"), False),
                        mavutil.mavlink.MAV_RESULT_FAILED: (_("Command failed"), False),
                    }

                    if msg.result in result_messages:
                        error_msg, success = result_messages[msg.result]
                        if not success:
                            logging_error(error_msg)
                        return success, error_msg

                    if msg.result == mavutil.mavlink.MAV_RESULT_IN_PROGRESS:
                        # Command is still in progress, continue waiting
                        if msg.progress is not None and msg.progress > 0:
                            logging_debug(_("Command in progress: %(progress)d%%"), {"progress": msg.progress})
                        continue

                    # Unknown result code
                    error_msg = _("Command acknowledgment with unknown result: %(result)d") % {"result": msg.result}
                    logging_error(error_msg)
                    return False, error_msg

                time_sleep(0.1)  # Sleep briefly to reduce CPU usage

            # Timeout occurred
            error_msg = _("Command acknowledgment timeout after %(timeout).1f seconds") % {"timeout": timeout}
            logging_error(error_msg)
            return False, error_msg

        except Exception as e:  # pylint: disable=broad-exception-caught
            error_msg = _("Failed to send command: %(error)s") % {"error": str(e)}
            logging_error(error_msg)
            return False, error_msg

    def __create_connection_with_retry(  # pylint: disable=too-many-arguments, too-many-positional-arguments, too-many-locals, too-many-branches
        self,
        progress_callback: Union[None, Callable[[int, int], None]],
        retries: int = 3,
        timeout: int = 5,
        baudrate: int = DEFAULT_BAUDRATE,
        log_errors: bool = True,
    ) -> str:
        """
        Attempts to create a connection to the flight controller with retries.

        This method attempts to establish a connection to the flight controller using the
        provided device connection string. It will retry the connection attempt up to the
        specified number of retries if the initial attempt fails. The method also supports
        a progress callback to report the progress of the connection attempt.

        Args:
            progress_callback (callable, optional): A callback function to report the progress
                                                    of the connection attempt. Default is None.
            retries (int, optional): The number of retries before giving up. Default is 3.
            timeout (int, optional): The timeout in seconds for each connection attempt. Default is 5.
            baudrate (int, optional): The baud rate for the connection. Default is DEFAULT_BAUDRATE.
            log_errors (bool): log errors.

        Returns:
            str: An error message if the connection fails after all retries, otherwise an empty string
                indicating a successful connection.

        """
        if self.comport is None or self.comport.device == "test":  # FIXME for testing only pylint: disable=fixme
            return ""
        if self.comport.device.startswith("udp") or self.comport.device.startswith("tcp"):
            logging_info(_("Will connect to %s"), self.comport.device)
        else:
            logging_info(_("Will connect to %s @ %u baud"), self.comport.device, baudrate)
        try:
            # Create the connection
            self.master = mavutil.mavlink_connection(
                device=self.comport.device,
                baud=baudrate,
                timeout=timeout,
                retries=retries,
                progress_callback=progress_callback,
            )
            logging_debug(_("Waiting for MAVLink heartbeats..."))
            if not self.master:
                msg = f"Failed to create mavlink connect to {self.comport.device}"
                raise ConnectionError(msg)
            # --- NEW: collect all vehicles detected within timeout ---
            start_time = time_time()
            detected_vehicles = {}  # (sysid, compid) -> last HEARTBEAT

            while time_time() - start_time < timeout:
                m = self.master.recv_match(type="HEARTBEAT", blocking=False)
                if m is None:
                    time_sleep(0.1)
                    continue
                sysid = m.get_srcSystem()
                compid = m.get_srcComponent()
                detected_vehicles[(sysid, compid)] = m
                logging_debug(_("Detected vehicle %u:%u (autopilot=%u, type=%u)"), sysid, compid, m.autopilot, m.type)

            if not detected_vehicles:
                return _("No MAVLink heartbeat received, connection failed.")

            for (sysid, compid), m in detected_vehicles.items():
                self.info.set_system_id_and_component_id(sysid, compid)
                logging_debug(
                    _("Connection established with systemID %d, componentID %d."), self.info.system_id, self.info.component_id
                )
                self.info.set_autopilot(m.autopilot)
                if self.info.is_supported:
                    msg = _("Autopilot type {self.info.autopilot}")
                    logging_info(msg.format(**locals()))
                    self.info.set_type(m.type)
                    msg = _("Vehicle type: {self.info.mav_type} running {self.info.vehicle_type} firmware")
                    logging_info(msg.format(**locals()))
                    break
                msg = _("Unsupported autopilot type {self.info.autopilot}")
                logging_info(msg.format(**locals()))

            if not self.info.is_supported:
                return _("No supported autopilots found")

            self.__request_banner()
            banner_msgs = self.__receive_banner_text()

            self.__request_message(mavutil.mavlink.MAVLINK_MSG_ID_AUTOPILOT_VERSION)
            m = self.master.recv_match(type="AUTOPILOT_VERSION", blocking=True, timeout=timeout)
            return self.__process_autopilot_version(m, banner_msgs)

        except (ConnectionError, SerialException, PermissionError, ConnectionRefusedError) as e:
            if log_errors:
                logging_warning(_("Connection failed: %s"), e)
                logging_error(_("Failed to connect after %d attempts."), retries)
            error_message = str(e)
            guidance = self.__get_connection_error_guidance(e, self.comport.device if self.comport else "")
            if guidance:
                error_message = f"{error_message}\n\n{guidance}"
            return error_message

    def __get_connection_error_guidance(self, error: Exception, device: str) -> str:
        """
        Provides guidance based on the type of connection error.

        Args:
            error (Exception): The exception that occurred during connection.
            device (str): The device path or connection string.

        Returns:
            str: Guidance message specific to the error type, or empty string if no specific guidance.

        """
        # Check for permission denied errors on Linux
        if isinstance(error, PermissionError) and os_name == "posix" and "/dev/" in device:
            return _(
                "Permission denied accessing the serial port. This is common on Linux systems.\n"
                "To fix this issue, add your user to the 'dialout' group with the following command:\n"
                "    sudo adduser $USER dialout\n"
                "Then log out and log back in for the changes to take effect."
            )

        # Add more specific guidance for other error types as needed

        return ""

    def __process_autopilot_version(self, m: MAVLink_autopilot_version_message, banner_msgs: list[str]) -> str:
        if m is None:
            return _(
                "No AUTOPILOT_VERSION MAVLink message received, connection failed.\n"
                "Only ArduPilot versions newer than 4.3.8 are supported.\n"
                "Make sure parameter SERIAL0_PROTOCOL is set to 2"
            )
        self.info.set_capabilities(m.capabilities)
        self.info.set_flight_sw_version(m.flight_sw_version)
        self.info.set_usb_vendor_and_product_ids(m.vendor_id, m.product_id)  # must be done before set_board_version()
        self.info.set_board_version(m.board_version)
        self.info.set_flight_custom_version(m.flight_custom_version)
        self.info.set_os_custom_version(m.os_custom_version)

        os_custom_version = ""
        os_custom_version_index = None
        for i, msg in enumerate(banner_msgs):
            if "ChibiOS:" in msg:
                os_custom_version = msg.split(" ")[1].strip()
                hash_len1 = max(7, len(os_custom_version) - 1)
                hash_len2 = max(7, len(self.info.os_custom_version) - 1)
                hash_len = min(hash_len1, hash_len2)
                if os_custom_version[:hash_len] != self.info.os_custom_version[:hash_len]:
                    logging_warning(
                        _("ChibiOS version mismatch: %s (BANNER) != % s (AUTOPILOT_VERSION)"),
                        os_custom_version,
                        self.info.os_custom_version,
                    )
                os_custom_version_index = i
                continue
            logging_debug("FC banner %s", msg)

        # the banner message after the ChibiOS one contains the FC type
        firmware_type = ""
        if os_custom_version_index is not None and os_custom_version_index + 1 < len(banner_msgs):
            firmware_type_banner_substrings = banner_msgs[os_custom_version_index + 1].split(" ")
            if len(firmware_type_banner_substrings) >= 3:
                firmware_type = firmware_type_banner_substrings[0]
        if firmware_type and firmware_type != self.info.firmware_type:
            logging_debug(
                _("FC firmware type mismatch: %s (BANNER) != %s (AUTOPILOT_VERSION)"), firmware_type, self.info.firmware_type
            )
            self.info.firmware_type = firmware_type  # force the one from the banner because it is more reliable
        return ""

    def download_params(
        self,
        progress_callback: Union[None, Callable[[int, int], None]] = None,
        parameter_values_filename: Optional[Path] = None,
        parameter_defaults_filename: Optional[Path] = None,
    ) -> tuple[dict[str, float], ParDict]:
        """
        Requests all flight controller parameters from a MAVLink connection.

        Args:
            progress_callback (Union[None, Callable[[int, int], None]]): A callback function to report download progress.
            parameter_values_filename (Optional[Path]): The filename to save the parameter values.
            parameter_defaults_filename (Optional[Path]): The filename to save the parameter defaults.

        Returns:
            dict[str, float]: A dictionary of flight controller parameters.
            ParDict: A dictionary of flight controller default parameters.

        """
        # FIXME this entire if statement is for testing only, remove it later pylint: disable=fixme
        if self.master is None and self.comport is not None and self.comport.device == "test":
            filename = "params.param"
            logging_warning(_("Testing active, will load all parameters from the %s file"), filename)
            par_dict_with_comments = ParDict.from_file(filename)
            return {k: v.value for k, v in par_dict_with_comments.items()}, ParDict()

        if self.master is None:
            return {}, ParDict()

        # Check if MAVFTP is supported
        comport_device = getattr(self.comport, "device", "")
        if self.info.is_mavftp_supported:
            logging_info(_("MAVFTP is supported by the %s flight controller"), comport_device)

            param_dict, default_param_dict = self._download_params_via_mavftp(
                progress_callback, parameter_values_filename, parameter_defaults_filename
            )
            if param_dict:
                return param_dict, default_param_dict

        logging_info(_("MAVFTP is not supported by the %s flight controller, fallback to MAVLink"), comport_device)
        return self._download_params_via_mavlink(progress_callback), ParDict()

    def _download_params_via_mavlink(
        self, progress_callback: Union[None, Callable[[int, int], None]] = None
    ) -> dict[str, float]:
        """
        Requests all flight controller parameters from a MAVLink connection.

        Gets parameters via PARAM_REQUEST_LIST and PARAM_VALUE messages

        Args:
            progress_callback (Union[None, Callable[[int, int], None]]): A callback function to report download progress.

        Returns:
            dict[str, float]: A dictionary of flight controller parameters.
            ParDict: A dictionary of flight controller default parameters.

        """
        comport_device = getattr(self.comport, "device", "")
        logging_debug(_("Will fetch all parameters from the %s flight controller"), comport_device)

        # Dictionary to store parameters
        parameters: dict[str, float] = {}

        # Request all parameters
        if self.master is None:
            return parameters

        self.master.mav.param_request_list_send(self.master.target_system, self.master.target_component)

        # Loop to receive all parameters
        while True:
            try:
                m = self.master.recv_match(type="PARAM_VALUE", blocking=True, timeout=10)
                if m is None:
                    break
                message = m.to_dict()
                param_id = message["param_id"]  # .decode("utf-8")
                param_value = message["param_value"]
                parameters[param_id] = param_value
                logging_debug(_("Received parameter: %s = %s"), param_id, param_value)
                # Call the progress callback with the current progress
                if progress_callback:
                    progress_callback(len(parameters), m.param_count)
                if m.param_count == len(parameters):
                    logging_debug(
                        _("Fetched %d parameter values from the %s flight controller"), m.param_count, comport_device
                    )
                    break
            except Exception as error:  # pylint: disable=broad-except
                logging_error(_("Error: %s"), error)
                break
        return parameters

    def _download_params_via_mavftp(
        self,
        progress_callback: Union[None, Callable[[int, int], None]] = None,
        parameter_values_filename: Optional[Path] = None,
        parameter_defaults_filename: Optional[Path] = None,
    ) -> tuple[dict[str, float], ParDict]:
        """
        Requests all flight controller parameters from a MAVLink connection.

        Gets parameters via MAVFTP protocol

        Args:
            progress_callback (Union[None, Callable[[int, int], None]]): A callback function to report download progress.
            parameter_values_filename (Optional[Path]): The filename to save the parameter values.
            parameter_defaults_filename (Optional[Path]): The filename to save the parameter defaults.

        Returns:
            dict[str, float]: A dictionary of flight controller parameters.
            ParDict: A dictionary of flight controller default parameters.

        """
        if self.master is None:
            return {}, ParDict()
        mavftp = MAVFTP(self.master, target_system=self.master.target_system, target_component=self.master.target_component)

        def get_params_progress_callback(completion: float) -> None:
            if progress_callback is not None and completion is not None:
                progress_callback(int(completion * 100), 100)

        complete_param_filename = str(parameter_values_filename) if parameter_values_filename else "complete.param"
        default_param_filename = str(parameter_defaults_filename) if parameter_defaults_filename else "00_default.param"
        mavftp.cmd_getparams([complete_param_filename, default_param_filename], progress_callback=get_params_progress_callback)
        ret = mavftp.process_ftp_reply("getparams", timeout=40)  #  on slow links it might take a long time
        pdict: dict[str, float] = {}
        defdict: ParDict = ParDict()

        # add a file sync operation to ensure the file is completely written
        time_sleep(0.3)
        if ret.error_code == 0:
            # load the parameters from the file
            par_dict = ParDict.from_file(complete_param_filename)
            pdict = {name: data.value for name, data in par_dict.items()}
            defdict = ParDict.from_file(default_param_filename)
        else:
            ret.display_message()

        return pdict, defdict

    def set_param(self, param_name: str, param_value: float) -> None:
        """
        Set a parameter on the flight controller.

        Args:
            param_name (str): The name of the parameter to set.
            param_value (float): The value to set the parameter to.

        """
        if self.master is None:  # FIXME for testing only pylint: disable=fixme
            return
        self.master.param_set_send(param_name, param_value)

    def fetch_param(self, param_name: str, timeout: int = 5) -> Optional[float]:
        """
        Fetch a parameter from the flight controller using MAVLink PARAM_REQUEST_READ message.

        Args:
            param_name (str): The name of the parameter to fetch.
            timeout (int): Timeout in seconds to wait for the response. Default is 5.

        Returns:
            float: The value of the parameter, or None if not found or timeout occurred.

        """
        if self.master is None:  # FIXME for testing only pylint: disable=fixme
            return None

        # Send PARAM_REQUEST_READ message
        self.master.mav.param_request_read_send(
            self.master.target_system,
            self.master.target_component,
            param_name.encode("utf-8"),
            -1,  # param_index: -1 means use param_id instead
        )

        # Wait for PARAM_VALUE response
        start_time = time_time()
        while time_time() - start_time < timeout:
            msg = self.master.recv_match(type="PARAM_VALUE", blocking=False)
            if msg is not None:
                # Check if this is the parameter we requested
                received_param_name = msg.param_id.rstrip("\x00")
                if received_param_name == param_name:
                    logging_debug(_("Received parameter: %s = %s"), param_name, msg.param_value)
                    return float(msg.param_value)
            time_sleep(0.01)  # Small sleep to prevent busy waiting

        raise TimeoutError(_("Timeout waiting for parameter %s") % param_name)

    def reset_all_parameters_to_default(self) -> tuple[bool, str]:
        """
        Reset all parameters to their factory default values.

        This function sends a MAV_CMD_PREFLIGHT_STORAGE command to reset all parameters
        to their factory defaults and waits for acknowledgment from the flight controller.
        The flight controller will need to be rebooted after this operation to apply the changes.

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        Note:
            After calling this method, the flight controller should be rebooted to
            apply the parameter reset. The reset operation will take effect only
            after the reboot.

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for parameter reset")
            logging_warning(error_msg)
            return False, error_msg

        # MAV_CMD_PREFLIGHT_STORAGE command
        # https://mavlink.io/en/messages/common.html#MAV_CMD_PREFLIGHT_STORAGE
        # param1 = 2: Erase all parameters
        success, error_msg = self._send_command_and_wait_ack(
            mavutil.mavlink.MAV_CMD_PREFLIGHT_STORAGE,
            param1=2,  # Storage action (2 = erase all parameters)
            param2=0,  # Parameter reset (0 = No parameter reset)
            param3=0,  # Mission reset (not used)
            param4=0,  # unused
            param5=0,  # unused
            param6=0,  # unused
            param7=0,  # unused
            timeout=10.0,  # Give more time for parameter reset
        )

        if success:
            logging_info(_("Parameter reset to defaults command confirmed by flight controller"))
        else:
            error_msg = _("Parameter reset command failed: %(error)s") % {"error": error_msg}
            logging_error(error_msg)

        return success, error_msg

    def reset_and_reconnect(
        self,
        reset_progress_callback: Union[None, Callable[[int, int], None]] = None,
        connection_progress_callback: Union[None, Callable[[int, int], None]] = None,
        extra_sleep_time: Optional[int] = None,
    ) -> str:
        """
        Reset the flight controller and reconnect.

        Args:
            reset_progress_callback: reset callback function
            connection_progress_callback: connection callback function
            extra_sleep_time (int, optional): The time in seconds to wait before reconnecting.

        """
        if self.master is None:  # FIXME for testing only pylint: disable=fixme
            return ""
        # Issue a reset
        self.master.reboot_autopilot()
        logging_info(_("Reset command sent to ArduPilot."))
        time_sleep(0.3)

        self.disconnect()

        current_step = 0

        if extra_sleep_time is None or extra_sleep_time < 0:
            extra_sleep_time = 0

        sleep_time = self.__reboot_time + extra_sleep_time

        while current_step != sleep_time:
            # Call the progress callback with the current progress
            if reset_progress_callback:
                reset_progress_callback(current_step, sleep_time)

            # Wait for sleep_time seconds
            time_sleep(1)
            current_step += 1

        # Call the progress callback with the current progress
        if reset_progress_callback:
            reset_progress_callback(current_step, sleep_time)

        # Reconnect to the flight controller
        return self.__create_connection_with_retry(connection_progress_callback, baudrate=self.__baudrate)

    @staticmethod
    def __list_serial_ports() -> list[serial.tools.list_ports_common.ListPortInfo]:
        """List all available serial ports."""
        comports = serial.tools.list_ports.comports()
        for port in comports:
            logging_debug("ComPort - %s, Description: %s", port.device, port.description)
        return comports  # type: ignore[no-any-return]

    # Motor Test Functionality

    def test_motor(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self, test_sequence_nr: int, motor_letters: str, motor_output_nr: int, throttle_percent: int, timeout_seconds: int
    ) -> tuple[bool, str]:
        """
        Test a specific motor.

        Args:
            test_sequence_nr: Motor test number, this is not the same as the output number!
            motor_letters: Motor letters (for logging purposes only)
            motor_output_nr: Motor output number (for logging purposes only)
            throttle_percent: Throttle percentage (0-100)
            timeout_seconds: Test duration in seconds

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for motor test")
            logging_error(error_msg)
            return False, error_msg

        # MAV_CMD_DO_MOTOR_TEST command
        # https://mavlink.io/en/messages/common.html#MAV_CMD_DO_MOTOR_TEST
        success, error_msg = self._send_command_and_wait_ack(
            mavutil.mavlink.MAV_CMD_DO_MOTOR_TEST,
            param1=test_sequence_nr + 1,  # motor test number, this is not the same as the output number!
            param2=mavutil.mavlink.MOTOR_TEST_THROTTLE_PERCENT,  # throttle type
            param3=throttle_percent,  # throttle value
            param4=timeout_seconds,  # timeout
            param5=0,  # motor count (0=test just the motor specified in param1)
            param6=0,  # test order (0=default/board order)
            param7=0,  # unused
        )

        if success:
            logging_info(
                _(
                    "Motor test command acknowledged: Motor %(seq)s on output %(output)d at %(throttle)d%% thrust"
                    " for %(duration)d seconds"
                ),
                {
                    "seq": motor_letters,
                    "output": motor_output_nr,
                    "throttle": throttle_percent,
                    "duration": timeout_seconds,
                },
            )
        else:
            error_msg = _("Motor test command failed: %(error)s") % {"error": error_msg}
            logging_error(error_msg)

        return success, error_msg

    def test_all_motors(self, nr_of_motors: int, throttle_percent: int, timeout_seconds: int) -> tuple[bool, str]:
        """
        Test all motors simultaneously.

        Args:
            nr_of_motors: Number of motors to test
            throttle_percent: Throttle percentage (0-100)
            timeout_seconds: Test duration in seconds

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for motor test")
            logging_error(error_msg)
            return False, error_msg

        for i in range(nr_of_motors):
            # MAV_CMD_DO_MOTOR_TEST command for all motors
            self.master.mav.command_long_send(
                self.master.target_system,
                self.master.target_component,
                mavutil.mavlink.MAV_CMD_DO_MOTOR_TEST,
                0,  # confirmation
                param1=i + 1,  # motor number (1-based)
                param2=mavutil.mavlink.MOTOR_TEST_THROTTLE_PERCENT,  # throttle type
                param3=throttle_percent,  # throttle value
                param4=timeout_seconds,  # timeout
                param5=0,  # motor count (0=all motors when param1=0)
                param6=0,  # test order (0=default/board order)
                param7=0,  # unused
            )
            time_sleep(0.01)  # to let the FC parse each command individually

        return True, ""

    def test_motors_in_sequence(
        self, start_motor: int, motor_count: int, throttle_percent: int, timeout_seconds: int
    ) -> tuple[bool, str]:
        """
        Test motors in sequence (A, B, C, D, etc.).

        Args:
            start_motor: The first motor to test (1-based index)
            motor_count: Number of motors to test in sequence
            throttle_percent: Throttle percentage (1-100)
            timeout_seconds: Test duration per motor in seconds

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for motor test")
            logging_error(error_msg)
            return False, error_msg

        # MAV_CMD_DO_MOTOR_TEST command for sequence test
        success, error_msg = self._send_command_and_wait_ack(
            mavutil.mavlink.MAV_CMD_DO_MOTOR_TEST,
            param1=start_motor,  # starting motor number (1-based)
            param2=mavutil.mavlink.MOTOR_TEST_THROTTLE_PERCENT,  # throttle type
            param3=throttle_percent,  # throttle value
            param4=timeout_seconds,  # timeout per motor
            param5=motor_count,  # number of motors to test in sequence
            param6=mavutil.mavlink.MOTOR_TEST_ORDER_SEQUENCE,  # test order (sequence)
            param7=0,  # unused
        )

        if success:
            logging_info(
                _("Sequential motor test command confirmed at %(throttle)d%% for %(duration)d seconds per motor"),
                {
                    "throttle": throttle_percent,
                    "duration": timeout_seconds,
                },
            )
        else:
            error_msg = _("Sequential motor test command failed: %(error)s") % {"error": error_msg}
            logging_error(error_msg)

        return success, error_msg

    def stop_all_motors(self) -> tuple[bool, str]:
        """
        Emergency stop for all motors.

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for motor stop")
            logging_error(error_msg)
            return False, error_msg

        # Send motor test command with 0% throttle to stop all motors
        success, error_msg = self._send_command_and_wait_ack(
            mavutil.mavlink.MAV_CMD_DO_MOTOR_TEST,
            param1=0,  # motor number (0 = all motors)
            param2=mavutil.mavlink.MOTOR_TEST_THROTTLE_PERCENT,  # throttle type
            param3=0,  # throttle value (0% = stop)
            param4=0,  # timeout (0 = immediate stop)
            param5=0,  # motor count (0 = all motors when param1=0)
            param6=0,  # test order (0 = default/board order)
            param7=0,  # unused
        )

        if success:
            logging_info(_("Motor stop command confirmed"))
        else:
            error_msg = _("Motor stop command failed: %(error)s") % {"error": error_msg}
            logging_error(error_msg)

        return success, error_msg

    def request_periodic_battery_status(self, interval_microseconds: int = 1000000) -> tuple[bool, str]:
        """
        Request periodic BATTERY_STATUS messages from the flight controller.

        Args:
            interval_microseconds: Message interval in microseconds (default: 1 second = 1,000,000 microseconds)

        Returns:
            tuple[bool, str]: (success, error_message) - success is True if command was acknowledged successfully,
                             error_message is empty string on success or contains error description on failure

        """
        if self.master is None:
            error_msg = _("No flight controller connection available for battery status request")
            logging_debug(error_msg)
            return False, error_msg

        # MAV_CMD_SET_MESSAGE_INTERVAL command to request periodic BATTERY_STATUS messages
        # https://mavlink.io/en/messages/common.html#MAV_CMD_SET_MESSAGE_INTERVAL
        success, error_msg = self._send_command_and_wait_ack(
            mavutil.mavlink.MAV_CMD_SET_MESSAGE_INTERVAL,
            param1=mavutil.mavlink.MAVLINK_MSG_ID_BATTERY_STATUS,  # message ID (BATTERY_STATUS)
            param2=interval_microseconds,  # interval in microseconds
            param3=0,  # unused
            param4=0,  # unused
            param5=0,  # unused
            param6=0,  # unused
            param7=0,  # unused
            timeout=0.8,  # shorter timeout for battery status requests
        )

        if success:
            logging_debug(
                _("Periodic BATTERY_STATUS messages confirmed every %(interval)d microseconds"),
                {"interval": interval_microseconds},
            )
        else:
            error_msg = _("Failed to request periodic battery status: %(error)s") % {"error": error_msg}
            logging_debug(error_msg)

        return success, error_msg

    def get_battery_status(self) -> tuple[Union[tuple[float, float], None], str]:
        """
        Get current battery voltage and current.

        Returns:
            tuple[Union[tuple[float, float], None], str]: ((voltage, current), error_message) -
                                                         voltage and current in volts and amps,
                                                         or None if not available with error message

        """
        if not self.fc_parameters or self.master is None:
            error_msg = _("No flight controller connection or parameters available")
            return None, error_msg

        # Check if battery monitoring is enabled
        if not self.is_battery_monitoring_enabled():
            error_msg = _("Battery monitoring is not enabled (BATT_MONITOR=0)")
            return None, error_msg

        try:
            # Try to get real telemetry data
            battery_status = self.master.recv_match(type="BATTERY_STATUS", blocking=False, timeout=0.3)
            if battery_status:
                # Convert from millivolts to volts, and centiamps to amps
                voltage = battery_status.voltages[0] / 1000.0 if battery_status.voltages[0] != -1 else 0.0
                current = battery_status.current_battery / 100.0 if battery_status.current_battery != -1 else 0.0
                self._last_battery_status = (voltage, current)
                self._last_battery_message_time = time_time()
                return (voltage, current), ""
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_debug(_("Failed to get battery status from telemetry: %(error)s"), {"error": str(e)})

        if self._last_battery_message_time and (time_time() - self._last_battery_message_time) < 3:
            # If we received a battery message recently, don't log an error
            return self._last_battery_status, ""
        self._last_battery_status = None
        error_msg = _("Battery status not available from telemetry")
        return None, error_msg

    def get_voltage_thresholds(self) -> tuple[float, float]:
        """
        Get battery voltage thresholds for motor testing safety.

        Returns:
            tuple[float, float]: (min_voltage, max_voltage) for safe motor testing

        """
        min_voltage = self.fc_parameters.get("BATT_ARM_VOLT", 0.0)
        max_voltage = self.fc_parameters.get("MOT_BAT_VOLT_MAX", 0.0)
        return (min_voltage, max_voltage)

    def is_battery_monitoring_enabled(self) -> bool:
        """
        Check if battery monitoring is enabled.

        Returns:
            bool: True if BATT_MONITOR != 0, False otherwise

        """
        return self.fc_parameters.get("BATT_MONITOR", 0) != 0

    def get_frame_info(self) -> tuple[int, int]:
        """
        Get frame class and frame type from flight controller parameters.

        Returns:
            tuple[int, int]: (frame_class, frame_type)

        """
        frame_class = int(self.fc_parameters.get("FRAME_CLASS", 1))  # Default to QUAD
        frame_type = int(self.fc_parameters.get("FRAME_TYPE", 1))  # Default to X
        return (frame_class, frame_type)

    @staticmethod
    def __list_network_ports() -> list[str]:
        """List all available network ports."""
        return ["tcp:127.0.0.1:5760", "udp:0.0.0.0:14550"]

    # pylint: disable=duplicate-code
    def __auto_detect_serial(self) -> list[mavutil.SerialPort]:
        preferred_ports = [
            "*FTDI*",
            "*3D*",
            "*USB_to_UART*",
            "*Ardu*",
            "*PX4*",
            "*Hex_*",
            "*ProfiCNC*",
            "*Holybro_*",
            "*mRo*",
            "*FMU*",
            "*Swift-Flyer*",
            "*Serial*",
            "*CubePilot*",
            "*Qiotek*",
        ]
        serial_list: list[mavutil.SerialPort] = [
            mavutil.SerialPort(device=connection[0], description=connection[1])
            for connection in self.__connection_tuples
            if connection[1] and "mavlink" in connection[1].lower()
        ]
        if len(serial_list) == 1:
            # selected automatically if unique
            return serial_list

        serial_list = mavutil.auto_detect_serial(preferred_list=preferred_ports)
        serial_list.sort(key=lambda x: x.device)

        # remove OTG2 ports for dual CDC
        if (
            len(serial_list) == 2
            and serial_list[0].device.startswith("/dev/serial/by-id")
            and serial_list[0].device[:-1] == serial_list[1].device[0:-1]
        ):
            serial_list.pop(1)

        return serial_list

    # pylint: enable=duplicate-code

    def get_connection_tuples(self) -> list[tuple[str, str]]:
        """Get all available connections."""
        return self.__connection_tuples

    def upload_file(
        self, local_filename: str, remote_filename: str, progress_callback: Union[None, Callable[[int, int], None]] = None
    ) -> bool:
        """Upload a file to the flight controller."""
        if self.master is None:
            return False
        mavftp = MAVFTP(self.master, target_system=self.master.target_system, target_component=self.master.target_component)

        def put_progress_callback(completion: float) -> None:
            if progress_callback is not None and completion is not None:
                progress_callback(int(completion * 100), 100)

        mavftp.cmd_put([local_filename, remote_filename], progress_callback=put_progress_callback)
        ret = mavftp.process_ftp_reply("CreateFile", timeout=10)
        if ret.error_code != 0:
            ret.display_message()
        return ret.error_code == 0

    def download_last_flight_log(
        self, local_filename: str, progress_callback: Union[None, Callable[[int, int], None]] = None
    ) -> bool:
        """Download the last flight log from the flight controller."""
        if self.master is None:
            error_msg = _("No flight controller connected")
            logging_error(error_msg)
            return False
        if not self.info.is_mavftp_supported:
            error_msg = _("MAVFTP is not supported by the flight controller")
            logging_error(error_msg)
            return False

        mavftp = MAVFTP(self.master, target_system=self.master.target_system, target_component=self.master.target_component)

        def get_progress_callback(completion: float) -> None:
            if progress_callback is not None and completion is not None:
                progress_callback(int(completion * 100), 100)

        try:
            # Try to get the last log number using different methods
            remote_filenumber = self._get_last_log_number(mavftp)
            if remote_filenumber is None:
                return False

            # We want the previous log, not the current one (which might be incomplete)
            # remote_filenumber -= 1
            # if remote_filenumber < 1:
            #     logging_error(_("No previous flight log available"))
            #     return False

            return self._download_log_file(mavftp, remote_filenumber, local_filename, get_progress_callback)

        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_error(_("Error during flight log download: %s"), str(e))
            return False

    def _get_last_log_number(self, mavftp: MAVFTP) -> Union[int, None]:
        """Get the last log number using multiple fallback methods."""
        # Method 1: Try to get LASTLOG.TXT
        log_number = self._get_log_number_from_lastlog_txt(mavftp)
        if log_number is not None:
            return log_number

        # Method 2: Try to list the logs directory and find the highest numbered log
        log_number = self._get_log_number_from_directory_listing(mavftp)
        if log_number is not None:
            return log_number

        # Method 3: Try common log numbers (scan backwards from a reasonable max)
        log_number = self._get_log_number_by_scanning(mavftp)
        if log_number is not None:
            return log_number

        logging_error(_("Could not determine the last log number using any method"))
        return None

    def _get_log_number_from_lastlog_txt(self, mavftp: MAVFTP) -> Union[int, None]:
        """Try to get the log number from LASTLOG.TXT file."""
        logging_info(_("Trying to get log number from LASTLOG.TXT"))
        try:
            temp_lastlog_file = "temp_lastlog.txt"
            mavftp.cmd_get(["/APM/LOGS/LASTLOG.TXT", temp_lastlog_file])
            ret = mavftp.process_ftp_reply("OpenFileRO", timeout=10)
            if ret.error_code != 0:
                logging_warning(_("LASTLOG.TXT not available, trying alternative methods"))
                return None

            return self._extract_log_number_from_file(temp_lastlog_file)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_warning(_("Failed to get log number from LASTLOG.TXT: %s"), str(e))
            return None

    def _get_log_number_from_directory_listing(self, _mavftp: MAVFTP) -> Union[int, None]:
        """Try to get the highest log number by listing the logs directory using MAVFTP."""
        logging_info(_("Trying to get log number from directory listing"))
        try:
            result = _mavftp.cmd_list(["/APM/LOGS/"])
            if not hasattr(result, "directory_listing") or not isinstance(result.directory_listing, dict):
                logging_error(_("No directory listing found in MAVFTPReturn"))
                return None
            highest = -1
            for name in result.directory_listing:
                # Typical log file names: 00000036.BIN, 00000037.BIN, etc.
                if name.endswith(".BIN") and name[:8].isdigit():
                    try:
                        log_num = int(name[:8])
                        highest = max(highest, log_num)
                    except ValueError:
                        continue
            if highest != -1:
                logging_info(_("Highest log number found: %d"), highest)
                return highest
            logging_error(_("No log files found in directory listing"))
            return None
        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_warning(_("Failed to get log number from directory listing: %s"), str(e))
            return None

    def _get_log_number_by_scanning(self, mavftp: MAVFTP) -> Union[int, None]:
        """Try to find the last log using binary search for efficiency."""
        logging_info(_("Trying to find log number using binary search"))
        try:
            # Binary search to find the highest log number
            low = 1
            high = 9999  # Reasonable upper bound for log numbers
            last_found = None

            while low <= high:
                mid = (low + high) // 2
                remote_filename = f"/APM/LOGS/{mid:08}.BIN"

                # Test if this log file exists
                temp_test_file = f"temp_test_{mid}.tmp"
                mavftp.cmd_get([remote_filename, temp_test_file])
                ret = mavftp.process_ftp_reply("OpenFileRO", timeout=5)  # Must be > idle_detection_time (3.7s)

                # Clean up the temp file if it was created
                if os.path.exists(temp_test_file):
                    os.remove(temp_test_file)

                if ret.error_code == 0:
                    # File exists, search in upper half
                    last_found = mid
                    low = mid + 1
                    logging_debug(_("Log %d exists, searching higher"), mid)
                else:
                    # File doesn't exist, search in lower half
                    high = mid - 1
                    logging_debug(_("Log %d doesn't exist, searching lower"), mid)

            if last_found is not None:
                logging_info(_("Found highest log number using binary search: %d"), last_found)
                return last_found

            logging_warning(_("No log files found using binary search"))
            return None

        except Exception as e:  # pylint: disable=broad-exception-caught
            logging_warning(_("Failed to scan for log numbers using binary search: %s"), str(e))
            return None

    def _download_log_file(
        self, mavftp: MAVFTP, remote_filenumber: int, local_filename: str, get_progress_callback: Callable
    ) -> bool:
        """Download the actual log file from the flight controller."""
        remote_filename = f"/APM/LOGS/{remote_filenumber:08}.BIN"
        logging_info(_("Downloading flight log %s to %s"), remote_filename, local_filename)

        # Download the actual log file
        mavftp.cmd_get([remote_filename, local_filename], progress_callback=get_progress_callback)
        ret = mavftp.process_ftp_reply("OpenFileRO", timeout=0)  # No timeout for large log files
        if ret.error_code != 0:
            logging_error(_("Failed to download flight log %s"), remote_filename)
            ret.display_message()
            return False

        logging_info(_("Successfully downloaded flight log to %s"), local_filename)
        return True

    def _extract_log_number_from_file(self, temp_lastlog_file: str) -> Union[int, None]:
        """Extract log number from LASTLOG.TXT file and clean up the temporary file."""
        try:
            with open(temp_lastlog_file, encoding="UTF-8") as file:
                file_contents = file.readline()
                return int(file_contents.strip())
        except (FileNotFoundError, ValueError) as e:
            logging_error(_("Could not extract last log file number from LASTLOG.TXT: %s"), e)
            return None
        finally:
            # Clean up the temporary file
            if os.path.exists(temp_lastlog_file):
                os.remove(temp_lastlog_file)

    @staticmethod
    def add_argparse_arguments(parser: ArgumentParser) -> ArgumentParser:
        parser.add_argument(
            "--baudrate",
            type=int,
            default=DEFAULT_BAUDRATE,
            help=_("MAVLink serial connection baudrate to the flight controller. Default is %(default)s"),
        )
        parser.add_argument(  # type: ignore[attr-defined]
            "--device",
            type=str,
            default="",
            help=_(
                "MAVLink connection string to the flight controller. "
                'If set to "none" no connection is made.'
                " Default is autodetection"
            ),
        ).completer = lambda **_: FlightController.__list_serial_ports()  # pyright: ignore[reportAttributeAccessIssue]
        parser.add_argument(
            "-r",
            "--reboot-time",
            type=int,
            min=5,
            max=50,
            action=CheckRange,
            default=DEFAULT_REBOOT_TIME,
            help=_("Flight controller reboot time. Default is %(default)s"),
        )
        return parser

    @property
    def comport_device(self) -> str:
        """Get the current self.comport.device string."""
        if self.comport is not None:
            return str(getattr(self.comport, "device", ""))
        return ""



#!/usr/bin/env python3

"""
Tests for the ParameterEditorWindow class.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar Lucas

SPDX-License-Identifier: GPL-3.0-or-later
"""

from typing import Any
from unittest.mock import ANY, MagicMock, patch

import pytest

from ardupilot_methodic_configurator.configuration_manager import ConfigurationManager
from ardupilot_methodic_configurator.data_model_par_dict import Par
from ardupilot_methodic_configurator.frontend_tkinter_parameter_editor import ParameterEditorWindow
from ardupilot_methodic_configurator.plugin_constants import PLUGIN_MOTOR_TEST

# pylint: disable=redefined-outer-name, too-many-arguments, too-many-positional-arguments, unused-argument, protected-access


@pytest.fixture
def mock_flight_controller() -> MagicMock:
    """Create a mock flight controller for testing."""
    mock_fc = MagicMock()
    mock_fc.fc_parameters = {"PARAM1": 1.0, "PARAM2": 2.0}
    return mock_fc


@pytest.fixture
def mock_local_filesystem() -> MagicMock:
    """Create a mock local filesystem for testing."""
    mock_fs = MagicMock()
    mock_fs.file_parameters = {"test_file.param": {"PARAM1": Par(1.0), "PARAM2": Par(2.0)}}
    return mock_fs


@pytest.fixture
def parameter_editor(root, mock_flight_controller, mock_local_filesystem) -> ParameterEditorWindow:
    """Create a ParameterEditorWindow instance for testing with real widgets in headless mode."""
    # Create the configuration manager for the test
    config_manager = ConfigurationManager("test_file.param", mock_flight_controller, mock_local_filesystem)

    # Create the object without calling __init__
    editor = ParameterEditorWindow.__new__(ParameterEditorWindow)

    # Create a mock for parameter_editor_table
    mock_parameter_editor_table = MagicMock()

    # Manually set required attributes for tests using real root
    editor.root = root
    editor.main_frame = MagicMock()  # Still mock the main frame to avoid complex UI setup
    editor.configuration_manager = config_manager  # Add the missing configuration_manager
    editor.flight_controller = mock_flight_controller
    editor.local_filesystem = mock_local_filesystem
    editor.at_least_one_changed_parameter_written = False
    editor.parameter_editor_table = mock_parameter_editor_table

    return editor


@pytest.fixture
def configured_parameter_editor(mock_flight_controller, mock_local_filesystem) -> ParameterEditorWindow:
    """Fixture providing a configured ParameterEditorWindow for business logic testing."""
    with (
        patch("tkinter.Tk"),
        patch.object(ParameterEditorWindow, "__init__", return_value=None),
        patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.logging_info"),
    ):
        # Create the configuration manager for the test
        config_manager = ConfigurationManager("test_file.param", mock_flight_controller, mock_local_filesystem)

        editor = ParameterEditorWindow.__new__(ParameterEditorWindow)
        editor.configuration_manager = config_manager
        editor.flight_controller = mock_flight_controller
        editor.local_filesystem = mock_local_filesystem
        return editor


class TestParameterEditorWindow:
    """Test cases for the ParameterEditorWindow class."""

    @patch("sys.exit")
    def test_should_copy_fc_values_to_file_no_auto_change(
        self, mock_exit: MagicMock, parameter_editor, mock_local_filesystem
    ) -> None:
        """Test that nothing happens when there is no auto_changed_by value."""
        # Mock _should_copy_fc_values_to_file to return False (no copy needed)
        parameter_editor.configuration_manager._should_copy_fc_values_to_file = MagicMock(return_value=(False, None, None))

        parameter_editor._ParameterEditorWindow__should_copy_fc_values_to_file("test_file.param")  # pylint: disable=protected-access

        parameter_editor.configuration_manager._should_copy_fc_values_to_file.assert_called_once_with("test_file.param")
        mock_exit.assert_not_called()

    @patch("sys.exit")
    @patch.object(ConfigurationManager, "handle_copy_fc_values_workflow", return_value=True)
    def test_user_can_successfully_copy_flight_controller_values_to_configuration_file(
        self, mock_workflow: MagicMock, mock_exit: MagicMock, parameter_editor, mock_local_filesystem
    ) -> None:
        """
        User can successfully copy flight controller parameter values to a configuration file.

        GIVEN: A user is prompted to copy FC values to a configuration file
        WHEN: The user chooses 'Yes' to proceed with copying
        THEN: The copy workflow is executed successfully
        AND: The application continues running without exiting
        """
        parameter_editor._ParameterEditorWindow__should_copy_fc_values_to_file("test_file.param")  # pylint: disable=protected-access

        # Verify the workflow method was called with correct arguments
        mock_workflow.assert_called_once()
        args, _ = mock_workflow.call_args
        assert args[0] == "test_file.param"  # selected_file
        assert callable(args[1])  # ask_user_choice callback
        assert callable(args[2])  # show_info callback

        mock_exit.assert_not_called()

    @patch("sys.exit")
    @patch.object(ConfigurationManager, "handle_copy_fc_values_workflow", return_value=None)
    def test_user_can_decline_copying_flight_controller_values_to_configuration_file(
        self, mock_workflow: MagicMock, mock_exit: MagicMock, parameter_editor, mock_local_filesystem
    ) -> None:
        """
        User can decline to copy flight controller parameter values to a configuration file.

        GIVEN: A user is prompted to copy FC values to a configuration file
        WHEN: The user chooses 'No' to decline copying
        THEN: The copy workflow is executed but no copying occurs
        AND: The application continues running without exiting
        """
        parameter_editor._ParameterEditorWindow__should_copy_fc_values_to_file("test_file.param")  # pylint: disable=protected-access

        # Verify the workflow method was called with correct arguments
        mock_workflow.assert_called_once()
        args, _ = mock_workflow.call_args
        assert args[0] == "test_file.param"  # selected_file
        assert callable(args[1])  # ask_user_choice callback
        assert callable(args[2])  # show_info callback

        mock_exit.assert_not_called()

    @patch("sys.exit")
    @patch.object(ConfigurationManager, "handle_copy_fc_values_workflow", return_value="close")
    def test_user_can_cancel_and_close_application_when_prompted_to_copy_flight_controller_values(
        self, mock_workflow: MagicMock, mock_exit: MagicMock, parameter_editor, mock_local_filesystem
    ) -> None:
        """
        User can cancel the copy operation and close the application when prompted.

        GIVEN: A user is prompted to copy FC values to a configuration file
        WHEN: The user chooses 'Close' to cancel and exit
        THEN: The copy workflow is executed but cancelled
        AND: The application exits gracefully
        """
        parameter_editor._ParameterEditorWindow__should_copy_fc_values_to_file("test_file.param")  # pylint: disable=protected-access

        # Verify the workflow method was called with correct arguments
        mock_workflow.assert_called_once()
        args, _ = mock_workflow.call_args
        assert args[0] == "test_file.param"  # selected_file
        assert callable(args[1])  # ask_user_choice callback
        assert callable(args[2])  # show_info callback

        mock_exit.assert_called_once_with(0)

    @patch("tkinter.Toplevel")
    @patch("tkinter.Label")
    @patch("tkinter.Frame")
    @patch("tkinter.Button")
    def test_dialog_creation(  # pylint: disable=too-many-locals
        self,
        mock_button: MagicMock,
        mock_frame: MagicMock,
        mock_label: MagicMock,
        mock_toplevel: MagicMock,
        parameter_editor,
        mock_local_filesystem,
        root,
    ) -> None:
        """Test the creation of the dialog with its components."""
        # Mock _should_copy_fc_values_to_file to return True with relevant params
        relevant_fc_params = {"PARAM1": 1.0, "PARAM2": 2.0}
        parameter_editor.configuration_manager._should_copy_fc_values_to_file = MagicMock(
            return_value=(True, relevant_fc_params, "External Tool")
        )

        # Setup the mock toplevel to better simulate dialog behavior
        mock_dialog = MagicMock()
        mock_toplevel.return_value = mock_dialog
        mock_dialog.result = [None]  # Initialize result list

        # Don't let the test exit
        with patch("sys.exit"):
            # Replace wait_window with a mock that doesn't block
            def fake_wait_window(*args: Any, **kwargs: Any) -> None:  # noqa: ANN401 # pylint: disable=unused-argument
                pass

            root.wait_window = MagicMock(side_effect=fake_wait_window)

            parameter_editor._ParameterEditorWindow__should_copy_fc_values_to_file("test_file.param")  # pylint: disable=protected-access

        # Verify dialog creation
        mock_toplevel.assert_called_once()

        # Check for label, buttons, and frame creation
        assert mock_label.call_count == 2  # message label and link label
        assert mock_button.call_count == 3  # Close, Yes, No buttons
        mock_frame.assert_called_once()

        # Verify message label creation
        message_label_call = mock_label.call_args_list[0]
        assert message_label_call[1]["text"].startswith("This configuration step requires external changes by: External Tool")
        assert message_label_call[1]["justify"] == "left"
        assert message_label_call[1]["padx"] == 20
        assert message_label_call[1]["pady"] == 10

        # Verify link label creation
        link_label_call = mock_label.call_args_list[1]
        assert link_label_call[1]["text"] == "Click here to open the Tuning Guide relevant Section"
        assert link_label_call[1]["fg"] == "blue"
        assert link_label_call[1]["cursor"] == "hand2"
        # Font should be a tuple with underline
        font_arg = link_label_call[1]["font"]
        assert isinstance(font_arg, tuple)
        assert len(font_arg) == 3
        assert font_arg[2] == "underline"

        # Verify link label is packed and bound correctly
        link_label_mock = mock_label.return_value
        # Both labels call pack, so check that the link label's pack call was made
        assert link_label_mock.pack.call_count == 2
        link_label_mock.pack.assert_any_call(pady=(0, 10))  # link label pack call
        link_label_mock.pack.assert_any_call(padx=10, pady=10)  # message label pack call
        link_label_mock.bind.assert_called_once_with("<Button-1>", ANY)  # lambda function, so use ANY

        # Verify buttons are created with correct text
        button_calls = mock_button.call_args_list
        assert button_calls[0][1]["text"] == "Close"
        assert button_calls[1][1]["text"] == "Yes"
        assert button_calls[2][1]["text"] == "No"

        # Verify button frame is packed correctly
        frame_mock = mock_frame.return_value
        frame_mock.pack.assert_called_once_with(pady=10)

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    def test_load_plugin_creates_motor_test_view_when_model_available(self, mock_factory, parameter_editor) -> None:
        """
        Test that __load_plugin creates plugin view using factory when data model is available.

        GIVEN: A plugin configuration for motor_test and a valid data model
        WHEN: __load_plugin is called
        THEN: Plugin factory is used to create the view with the correct parameters
        """
        # Mock the factory methods
        mock_factory.is_registered.return_value = True
        mock_plugin_view = MagicMock()
        mock_plugin_view.pack = MagicMock()
        mock_factory.create.return_value = mock_plugin_view

        # Mock the configuration manager to return a valid model
        mock_model = MagicMock()
        with patch.object(parameter_editor.configuration_manager, "create_plugin_data_model", return_value=mock_model):
            # Create a mock parent frame
            mock_parent_frame = MagicMock()

            # Call the method
            plugin_config = {"name": "motor_test"}
            parameter_editor._ParameterEditorWindow__load_plugin(mock_parent_frame, plugin_config)  # pylint: disable=protected-access

            # Verify factory was used correctly
            mock_factory.is_registered.assert_called_once_with("motor_test")
            mock_factory.create.assert_called_once_with("motor_test", mock_parent_frame, mock_model, parameter_editor)
            mock_plugin_view.pack.assert_called_once()

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    def test_load_plugin_shows_error_message_when_model_creation_fails(self, mock_factory, parameter_editor) -> None:
        """
        Test that __load_plugin shows error message when data model creation fails.

        GIVEN: A plugin configuration for motor_test but model creation returns None
        WHEN: __load_plugin is called
        THEN: An error label is displayed and plugin is not created
        """
        # Mock the factory methods
        mock_factory.is_registered.return_value = True

        # Mock the configuration manager to return None (model creation failed)
        with patch.object(parameter_editor.configuration_manager, "create_plugin_data_model", return_value=None):
            # Create a mock parent frame
            mock_parent_frame = MagicMock()

            # Call the method
            plugin_config = {"name": "motor_test"}
            parameter_editor._ParameterEditorWindow__load_plugin(mock_parent_frame, plugin_config)  # pylint: disable=protected-access

            # Verify factory create was not called
            mock_factory.create.assert_not_called()

            # Verify error label was created and packed
            mock_parent_frame.children = []  # Simulate ttk.Frame behavior
            # The actual label creation happens inside the method

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ttk.Label")
    def test_load_plugin_shows_error_when_model_creation_returns_none(
        self, mock_label, mock_factory, parameter_editor
    ) -> None:
        """
        Test that __load_plugin shows error message when create_plugin_data_model returns None.

        GIVEN: create_plugin_data_model returns None (FC not connected)
        WHEN: __load_plugin is called for motor_test
        THEN: An error label is displayed
        """
        # Mock the factory
        mock_factory.is_registered.return_value = True

        # Mock the configuration manager to return None
        with patch.object(parameter_editor.configuration_manager, "create_plugin_data_model", return_value=None):
            # Create a mock parent frame
            mock_parent_frame = MagicMock()

            # Call the method
            plugin_config = {"name": "motor_test"}
            parameter_editor._ParameterEditorWindow__load_plugin(mock_parent_frame, plugin_config)  # pylint: disable=protected-access

            # Verify error label was created
            mock_label.assert_called_with(mock_parent_frame, text="Plugin requires flight controller connection")
            mock_label.return_value.pack.assert_called_once()

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    def test_load_plugin_shows_error_for_unknown_plugin(self, mock_factory, parameter_editor) -> None:
        """
        Test that __load_plugin shows error message for unknown plugin names.

        GIVEN: A plugin configuration with an unknown plugin name
        WHEN: __load_plugin is called
        THEN: An error label is displayed
        """
        # Mock factory to report plugin not registered
        mock_factory.is_registered.return_value = False

        # Create a mock parent frame
        mock_parent_frame = MagicMock()

        # Call the method with unknown plugin
        plugin_config = {"name": "unknown_plugin"}
        with patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ttk.Label") as mock_label:
            parameter_editor._ParameterEditorWindow__load_plugin(mock_parent_frame, plugin_config)  # pylint: disable=protected-access

            # Verify error label was created with translated text
            mock_label.assert_called_once()
            call_args = mock_label.call_args
            assert call_args[0][0] == mock_parent_frame
            assert "unknown_plugin" in call_args[1]["text"]
            mock_label.return_value.pack.assert_called_once()

    def test_update_plugin_layout_switches_from_plugin_to_no_plugin(self, parameter_editor) -> None:
        """
        Test switching from a plugin file to a non-plugin file properly cleans up.

        GIVEN: A parameter editor with an active plugin layout
        WHEN: __update_plugin_layout is called with None (no plugin)
        THEN: The plugin layout is destroyed and normal layout is restored
        """
        # Setup: Set current plugin to simulate having a plugin active
        parameter_editor.current_plugin = {"name": PLUGIN_MOTOR_TEST, "placement": "left"}
        parameter_editor.current_plugin_view = None  # Initialize attribute needed by cleanup
        parameter_editor.parameter_area_paned = None  # Initialize attribute needed by cleanup

        # Mock the necessary attributes
        mock_container = MagicMock()
        mock_table = MagicMock()
        parameter_editor.parameter_area_container = mock_container
        parameter_editor.parameter_editor_table = mock_table
        parameter_editor.main_frame = MagicMock()

        with (
            patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ttk.Frame") as mock_frame,
            patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ParameterEditorTable"),
        ):
            # Call with None to simulate switching to non-plugin file
            parameter_editor._ParameterEditorWindow__update_plugin_layout(None)  # pylint: disable=protected-access

            # Verify cleanup happened - container is destroyed (not individual widgets)
            mock_container.destroy.assert_called_once()

            # Verify new container was created
            mock_frame.assert_called()

            # Verify current_plugin was updated to None
            assert parameter_editor.current_plugin is None

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    def test_load_plugin_handles_factory_exception(self, mock_factory, parameter_editor) -> None:
        """
        Test that __load_plugin handles exceptions from the plugin factory gracefully.

        GIVEN: A registered plugin but factory.create raises an exception
        WHEN: __load_plugin is called
        THEN: An error message is displayed and logged
        """
        # Mock the factory methods
        mock_factory.is_registered.return_value = True
        mock_factory.create.side_effect = TypeError("Invalid arguments")

        # Mock the configuration manager to return a valid model
        mock_model = MagicMock()
        with patch.object(parameter_editor.configuration_manager, "create_plugin_data_model", return_value=mock_model):
            # Create a mock parent frame
            mock_parent_frame = MagicMock()

            # Call the method
            plugin_config = {"name": "motor_test"}
            with patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ttk.Label") as mock_label:
                parameter_editor._ParameterEditorWindow__load_plugin(mock_parent_frame, plugin_config)  # pylint: disable=protected-access

                # Verify error label was created with proper styling
                assert mock_label.called
                call_args = mock_label.call_args
                assert "foreground" in call_args[1]
                assert call_args[1]["foreground"] == "red"

    @patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.plugin_factory")
    def test_switching_between_two_plugin_files_calls_lifecycle_hooks(self, mock_factory, parameter_editor) -> None:
        """
        Test that switching between two different plugin files calls lifecycle hooks properly.

        GIVEN: Two different plugin configurations
        WHEN: Switching from one plugin file to another
        THEN: on_deactivate is called on the first plugin and on_activate on the second
        """
        # Setup: Mock factory
        mock_factory.is_registered.return_value = True

        # Create two mock plugin views with lifecycle hooks
        mock_plugin_1 = MagicMock()
        mock_plugin_1.on_activate = MagicMock()
        mock_plugin_1.on_deactivate = MagicMock()
        mock_plugin_1.pack = MagicMock()
        mock_plugin_1.destroy = MagicMock()

        mock_plugin_2 = MagicMock()
        mock_plugin_2.on_activate = MagicMock()
        mock_plugin_2.on_deactivate = MagicMock()
        mock_plugin_2.pack = MagicMock()
        mock_plugin_2.destroy = MagicMock()

        mock_factory.create.side_effect = [mock_plugin_1, mock_plugin_2]

        # Mock the configuration manager to return a valid model
        mock_model = MagicMock()
        with (
            patch.object(parameter_editor.configuration_manager, "create_plugin_data_model", return_value=mock_model),
            patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ttk.Frame"),
            patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.tk.PanedWindow"),
            patch("ardupilot_methodic_configurator.frontend_tkinter_parameter_editor.ParameterEditorTable"),
        ):
            # Setup initial state
            parameter_editor.current_plugin = None
            parameter_editor.current_plugin_view = None
            parameter_editor.parameter_area_paned = None
            parameter_editor.parameter_area_container = MagicMock()
            parameter_editor.parameter_editor_table = MagicMock()
            parameter_editor.main_frame = MagicMock()

            # First plugin activation
            plugin_config_1 = {"name": "motor_test", "placement": "left"}
            parameter_editor._ParameterEditorWindow__update_plugin_layout(plugin_config_1)  # pylint: disable=protected-access

            # Verify first plugin was activated
            mock_plugin_1.on_activate.assert_called_once()
            mock_plugin_1.on_deactivate.assert_not_called()

            # Second plugin activation
            plugin_config_2 = {"name": "other_plugin", "placement": "top"}
            parameter_editor._ParameterEditorWindow__update_plugin_layout(plugin_config_2)  # pylint: disable=protected-access

            # Verify first plugin was deactivated and destroyed
            mock_plugin_1.on_deactivate.assert_called_once()
            mock_plugin_1.destroy.assert_called_once()

            # Verify second plugin was activated
            mock_plugin_2.on_activate.assert_called_once()



"""
Data model for vehicle components JSON schema.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from logging import debug as logging_debug
from logging import error as logging_error
from typing import Any

from ardupilot_methodic_configurator import _


class VehicleComponentsJsonSchema:
    """Vehicle components JSON schema Business logic."""

    def __init__(self, schema: dict[str, Any]) -> None:
        self.schema = schema

    def get_all_value_datatypes(self) -> dict[str, Any]:
        """
        Get a dictionary of all value data types used in the vehicle_components_dict as defined in the schema.

        The keys are the vehicle_components_dict keys and their values are the actual Python datatype

        :return: A nested dictionary mapping data names to their Python datatypes.
        """
        value_datatypes: dict[str, Any] = {}

        # Start with the Components node since all our editable items are under it
        components_schema = self.schema.get("properties", {}).get("Components", {})
        if not components_schema:
            return {}

        # Traverse all component types (Flight Controller, Frame, etc.)
        for component_type, component_schema in components_schema.get("properties", {}).items():
            value_datatypes[component_type] = {}
            self._extract_datatypes_from_component(component_schema, value_datatypes[component_type])

        return value_datatypes

    def _extract_datatypes_from_component(self, component_schema: dict[str, Any], target_dict: dict[str, Any]) -> None:
        """Extract datatypes from a component schema recursively."""
        # Resolve reference if present
        resolved_schema = self._resolve_schema_reference(component_schema)

        # Check direct properties
        if "properties" in resolved_schema:
            for prop_name, prop_schema in resolved_schema["properties"].items():
                self._extract_datatypes_from_property(prop_schema, prop_name, target_dict)

        # Check allOf constructs
        if "allOf" in resolved_schema:
            for allof_item in resolved_schema["allOf"]:
                self._extract_datatypes_from_component(allof_item, target_dict)

    def _extract_datatypes_from_property(
        self, prop_schema: dict[str, Any], prop_name: str, target_dict: dict[str, Any]
    ) -> None:
        """Extract datatype from a property schema."""
        # Resolve reference if present
        resolved_schema = self._resolve_schema_reference(prop_schema)

        # If this property has nested properties, create a nested dict and recurse
        if "properties" in resolved_schema:
            target_dict[prop_name] = {}
            for nested_prop_name, nested_prop_schema in resolved_schema["properties"].items():
                self._extract_datatypes_from_property(nested_prop_schema, nested_prop_name, target_dict[prop_name])

        # Handle allOf constructs in properties
        elif "allOf" in resolved_schema:
            target_dict[prop_name] = {}
            for allof_item in resolved_schema["allOf"]:
                self._extract_datatypes_from_property(allof_item, prop_name, target_dict)

        # If this property has a direct type, convert to Python type and record it as a leaf value
        elif "type" in resolved_schema:
            target_dict[prop_name] = self._json_type_to_python_type(resolved_schema["type"])

        # If no type or properties found, create empty dict as placeholder
        else:
            target_dict[prop_name] = {}

    def _json_type_to_python_type(self, json_type: str) -> type:
        """
        Convert JSON Schema type string to actual Python type.

        :param json_type: JSON Schema type string
        :return: Corresponding Python type
        """
        type_mapping = {
            "string": str,
            "number": float,
            "integer": int,
            "boolean": bool,
            "array": list,
            "object": dict,
            "null": type(None),
        }

        return type_mapping.get(json_type, str)  # Default to str if unknown type

    def modify_schema_for_mcu_series(self, is_optional: bool) -> None:
        """
        Modify the schema to set x-is-optional property for MCU Series field based on its value.

        Dynamically updates the schema to make the MCU Series field optional

        Args:
            is_optional: If it should be marked as optional

        """
        if not self.schema:  # Still None after loading
            logging_error(_("Cannot modify schema: Schema could not be loaded"))
            return

        # Navigate to the MCU Series field in the schema
        flight_controller_def = self.schema.get("definitions", {}).get("flightController", {})
        flight_controller_properties = None

        # Check in allOf construct
        if "allOf" in flight_controller_def:
            for item in flight_controller_def["allOf"]:
                if "properties" in item and "Specifications" in item["properties"]:
                    flight_controller_properties = item
                    break

        if not flight_controller_properties:
            logging_error(_("Could not find Specifications in flight controller schema"))
            return

        specifications = flight_controller_properties["properties"]["Specifications"]

        if "properties" in specifications and "MCU Series" in specifications["properties"]:
            try:
                mcu_series_field = specifications["properties"]["MCU Series"]

                if is_optional:
                    mcu_series_field["x-is-optional"] = True
                else:
                    # Remove x-is-optional if it exists
                    mcu_series_field.pop("x-is-optional", None)

                logging_debug(
                    _("Modified schema: MCU Series x-is-optional=%s to value=%u"),
                    mcu_series_field.get("x-is-optional", False),
                    is_optional,
                )
            except Exception as err:  # pylint: disable=broad-exception-caught
                logging_error(_("Error modifying schema for MCU Series: %s"), str(err))

    def get_component_property_description(self, path: tuple[str, ...]) -> tuple[str, bool]:
        """
        Get description and optional status from schema using a component path.

        Args:
            path (tuple): The path to the component in the JSON data.

        Returns:
            tuple[str, bool]: A tuple containing (description, is_optional),
                            where is_optional defaults to False if not specified.

        """
        if not path or len(path) == 0:
            return ("", False)

        try:
            # Start with the Components node since all our editable items are under it
            current = self.schema.get("properties", {}).get("Components", {})

            # Handle different path scenarios
            if len(path) == 1:
                return self._get_top_level_component_description(current, path[0])
            if len(path) == 3 and path[1] == "Product":
                return self._get_product_field_description(path[2])
            return self._get_nested_property_description(current, path)
        except Exception as _e:  # pylint: disable=broad-exception-caught
            msg = _("Exception occurred in get_component_property_description: {}").format(str(_e))
            logging_error(msg)
            return ("", False)

    def _get_component_schema_property_description(self, component_info: dict[str, Any]) -> tuple[str, bool]:
        """Get description and optional status from a component schema property."""
        description = component_info.get("description", "")
        is_optional = component_info.get("x-is-optional", False)
        return (description, is_optional)

    def _get_top_level_component_description(self, current_schema: dict[str, Any], component_type: str) -> tuple[str, bool]:
        """Get description for top-level components like 'Flight Controller'."""
        if component_type in current_schema.get("properties", {}):
            return self._get_component_schema_property_description(current_schema["properties"][component_type])
        return ("", False)

    def _get_product_field_description(self, field_name: str) -> tuple[str, bool]:
        """Get description for product fields (Manufacturer, Model, etc.)."""
        product_def = self.schema.get("definitions", {}).get("product", {})
        if "properties" in product_def and field_name in product_def["properties"]:
            return self._get_component_schema_property_description(product_def["properties"][field_name])
        return ("", False)

    def _get_nested_property_description(self, current_schema: dict[str, Any], path: tuple[str, ...]) -> tuple[str, bool]:
        """Get description for nested properties in the component structure."""
        # Get the component type (e.g., "Flight Controller")
        component_type = path[0]

        # Navigate to the specific component type
        if component_type in current_schema.get("properties", {}):
            current = current_schema["properties"][component_type]
        else:
            return ("", False)

        # Resolve reference if present
        current = self._resolve_schema_reference(current)

        # Handle requests for section fields (Product, Firmware, Specifications, etc.)
        if len(path) == 2:
            return self._get_section_field_description(current, path[1])

        # For deeper nested fields, navigate through the path
        return self._traverse_nested_path(current, path[1:])

    def _resolve_schema_reference(self, schema_obj: dict[str, Any]) -> dict[str, Any]:
        """Resolve a schema reference in the form of {"$ref": "#/path/to/definition"}."""
        if "$ref" in schema_obj:
            ref_path = schema_obj["$ref"].replace("#/", "").split("/")
            ref_obj = self.schema
            for ref_part in ref_path:
                ref_obj = ref_obj.get(ref_part, {})
            return ref_obj
        return schema_obj

    def _get_section_field_description(self, current: dict[str, Any], section: str) -> tuple[str, bool]:
        """Get description for section fields like Product, Firmware, etc."""
        # First check in direct properties
        if "properties" in current and section in current["properties"]:
            return self._get_component_schema_property_description(current["properties"][section])

        # Then check in allOf constructs
        if "allOf" in current:
            for allof_item in current["allOf"]:
                # Handle reference in allOf item
                if "$ref" in allof_item:
                    ref_obj = self._resolve_schema_reference(allof_item)

                    if "properties" in ref_obj and section in ref_obj["properties"]:
                        return self._get_component_schema_property_description(ref_obj["properties"][section])

                # Direct properties check in this allOf item
                elif "properties" in allof_item and section in allof_item["properties"]:
                    return self._get_component_schema_property_description(allof_item["properties"][section])

        # If not found, return empty with default optional status
        return ("", False)

    def _traverse_nested_path(self, current: dict[str, Any], path_parts: tuple[str, ...]) -> tuple[str, bool]:
        """Traverse a nested path in the schema to find a property description."""
        for part in path_parts:
            found = False

            # Check strategies in order: direct properties, allOf, and references
            found, current = self._check_direct_properties(current, part)
            if not found:
                found, current = self._check_allof_constructs(current, part)
            if not found:
                found, current = self._check_references(current, part)

            # If not found after all checks, return empty
            if not found:
                return ("", False)

            # If we found a $ref in the current object, resolve it
            current = self._resolve_schema_reference(current)

        # Return the description and optional status of the final object
        return self._get_component_schema_property_description(current)

    def _check_direct_properties(self, schema_obj: dict[str, Any], property_name: str) -> tuple[bool, dict]:
        """Check if property exists directly in schema's properties."""
        if "properties" in schema_obj and property_name in schema_obj["properties"]:
            return True, schema_obj["properties"][property_name]
        return False, schema_obj

    def _check_allof_constructs(self, schema_obj: dict[str, Any], property_name: str) -> tuple[bool, dict]:
        """Check if property exists in any allOf constructs."""
        if "allOf" in schema_obj:
            for allof_item in schema_obj["allOf"]:
                # Handle reference in allOf item
                if "$ref" in allof_item:
                    ref_obj = self._resolve_schema_reference(allof_item)

                    if "properties" in ref_obj and property_name in ref_obj["properties"]:
                        return True, ref_obj["properties"][property_name]

                # Direct check in this allOf item
                elif "properties" in allof_item and property_name in allof_item["properties"]:
                    return True, allof_item["properties"][property_name]

        return False, schema_obj

    def _check_references(self, schema_obj: dict[str, Any], property_name: str) -> tuple[bool, dict]:
        """Check if property exists in referenced schema object."""
        if "$ref" in schema_obj:
            ref_obj = self._resolve_schema_reference(schema_obj)

            # Look in the resolved reference direct properties
            if "properties" in ref_obj and property_name in ref_obj["properties"]:
                return True, ref_obj["properties"][property_name]

            # Look in allOf constructs in the reference
            if "allOf" in ref_obj:
                for allof_item in ref_obj["allOf"]:
                    if "properties" in allof_item and property_name in allof_item["properties"]:
                        return True, allof_item["properties"][property_name]

        return False, schema_obj



#!/usr/bin/env python3

"""
BDD-style tests for the backend_filesystem_program_settings.py file.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import json
import platform
from os import path as os_path
from unittest.mock import MagicMock, mock_open, patch

import pytest

from ardupilot_methodic_configurator.backend_filesystem_program_settings import ProgramSettings

# pylint: disable=too-many-lines,protected-access,redefined-outer-name,too-few-public-methods


@pytest.fixture
def mock_file_operations() -> MagicMock:
    """Fixture providing mocked filesystem operations for path testing."""
    mock_ops = MagicMock()
    mock_ops.dirname.return_value = "/mock/app/dir"
    mock_ops.abspath.return_value = "/mock/app/dir/file"
    mock_ops.join.return_value = "/mock/app/dir/images/file.png"
    return mock_ops


@pytest.fixture
def mock_user_config() -> dict[str, str]:
    """Fixture providing mock user configuration directory paths."""
    return {
        "config_dir": "/mock/user/config",
        "settings_file": "/mock/user/config/settings.json",
    }


@pytest.fixture
def sample_program_settings() -> dict:
    """Fixture providing realistic program settings data for testing."""
    return {
        "Format version": 1,
        "directory_selection": {"template_dir": "/mock/template"},
        "display_usage_popup": {"component_editor": False, "parameter_editor": True},
        "auto_open_doc_in_browser": False,
        "motor_test_duration": 5.0,
        "motor_test_throttle_pct": 10,
    }


class TestApplicationResourcePaths:
    """Test user access to application resource file paths."""

    def test_user_can_get_application_icon_filepath(self) -> None:
        """
        User can retrieve the application icon filepath for GUI display.

        GIVEN: A user needs to display the application icon in the GUI
        WHEN: User requests the application icon filepath
        THEN: The correct path to ArduPilot_icon.png should be returned
        AND: The path should be properly constructed using importlib.resources
        """
        # Arrange: Mock the entire method to return expected path
        expected_path = "/mock/path/images/ArduPilot_icon.png"
        with patch.object(ProgramSettings, "application_icon_filepath", return_value=expected_path) as mock_method:
            # Act: Get application icon filepath
            result = ProgramSettings.application_icon_filepath()

            # Assert: Correct filepath is returned
            assert result == expected_path
            mock_method.assert_called_once()

    def test_user_can_get_application_logo_filepath(self) -> None:
        """
        User can retrieve the application logo filepath for display in dialogs.

        GIVEN: A user needs to display the application logo in about dialog
        WHEN: User requests the application logo filepath
        THEN: The correct path to ArduPilot_logo.png should be returned
        AND: The path should be properly constructed using importlib.resources
        """
        # Arrange: Mock the entire method to return expected path
        expected_path = "/mock/path/images/ArduPilot_logo.png"
        with patch.object(ProgramSettings, "application_logo_filepath", return_value=expected_path) as mock_method:
            # Act: Get application logo filepath
            result = ProgramSettings.application_logo_filepath()

            # Assert: Correct filepath is returned
            assert result == expected_path
            mock_method.assert_called_once()

    def test_user_can_get_icon_filepath_using_importlib_resources(self) -> None:
        """
        User can retrieve icon filepath using modern importlib.resources method.

        GIVEN: Python 3.9+ with importlib.resources.files available
        WHEN: User requests the application icon filepath
        THEN: The path should be retrieved using importlib.resources
        AND: The path should exist and end with ArduPilot_icon.png
        """
        # Act: Get application icon filepath (uses importlib.resources in Python 3.9+)
        result = ProgramSettings.application_icon_filepath()

        # Assert: Path is valid and ends with expected filename
        assert result.endswith("ArduPilot_icon.png")
        assert "images" in result
        # Path should exist when running from source or installed package
        assert os_path.exists(result), f"Icon file should exist at {result}"

    def test_user_can_get_logo_filepath_using_importlib_resources(self) -> None:
        """
        User can retrieve logo filepath using modern importlib.resources method.

        GIVEN: Python 3.9+ with importlib.resources.files available
        WHEN: User requests the application logo filepath
        THEN: The path should be retrieved using importlib.resources
        AND: The path should exist and end with ArduPilot_logo.png
        """
        # Act: Get application logo filepath (uses importlib.resources in Python 3.9+)
        result = ProgramSettings.application_logo_filepath()

        # Assert: Path is valid and ends with expected filename
        assert result.endswith("ArduPilot_logo.png")
        assert "images" in result
        # Path should exist when running from source or installed package
        assert os_path.exists(result), f"Logo file should exist at {result}"


class TestDirectoryManagement:
    """Test user directory creation and validation workflows."""

    def test_user_can_create_program_settings_instance(self) -> None:
        """
        User can instantiate ProgramSettings class for configuration management.

        GIVEN: A user needs to work with program settings
        WHEN: User creates a ProgramSettings instance
        THEN: The instance should be created successfully
        AND: Should be ready for configuration operations
        """
        # Arrange & Act: Create ProgramSettings instance
        settings = ProgramSettings()

        # Assert: Instance created successfully
        assert settings is not None
        assert isinstance(settings, ProgramSettings)

    def test_user_successfully_creates_new_vehicle_directory(self) -> None:
        """
        User can successfully create a new vehicle directory when path is available.

        GIVEN: A user wants to create a new vehicle directory
        WHEN: The target directory path does not exist and creation succeeds
        THEN: Directory should be created successfully
        AND: Empty string should be returned indicating success
        """
        # Arrange: Mock successful directory creation
        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_makedirs") as mock_makedirs,
        ):
            # Act: Create new vehicle directory
            result = ProgramSettings.create_new_vehicle_dir("/mock/new/vehicle/dir")

            # Assert: Directory created and success indicated
            mock_makedirs.assert_called_once_with("/mock/new/vehicle/dir", exist_ok=True)
            assert result == ""

    def test_user_gets_error_when_creating_existing_directory(self) -> None:
        """
        User receives error message when trying to create directory that already exists.

        GIVEN: A user attempts to create a new vehicle directory
        WHEN: The target directory already exists on the filesystem
        THEN: An error message should be returned instead of creating the directory
        """
        # Arrange: Mock directory already exists
        with patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True):
            # Act: Attempt to create existing directory
            result = ProgramSettings.create_new_vehicle_dir("/mock/existing/dir")

            # Assert: Error message is returned
            assert result != ""
            assert isinstance(result, str)

    def test_user_gets_error_when_directory_creation_fails(self) -> None:
        """
        User receives informative error when directory creation fails due to system error.

        GIVEN: A user attempts to create a new vehicle directory
        WHEN: Directory creation fails due to filesystem error (permissions, disk space, etc.)
        THEN: An error message containing the system error should be returned
        AND: The error should be logged for troubleshooting
        """
        # Arrange: Mock directory creation failure
        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False),
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.os_makedirs",
                side_effect=OSError("Permission denied"),
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_error") as mock_log_error,
        ):
            # Act: Attempt to create directory
            result = ProgramSettings.create_new_vehicle_dir("/mock/protected/dir")

            # Assert: Error is logged and message contains error details
            mock_log_error.assert_called_once()
            assert "Permission denied" in result

    def test_user_can_validate_directory_names_correctly(self) -> None:
        """
        User receives correct validation results for various directory name formats.

        GIVEN: A user needs to validate potential directory names
        WHEN: User checks various directory name formats
        THEN: Validation should correctly identify valid and invalid names
        AND: Platform-specific path separators should be handled appropriately
        """
        # Arrange & Act & Assert: Test various directory name patterns

        # Valid names should pass validation
        assert ProgramSettings.valid_directory_name("valid_dir_name-123") is True
        assert ProgramSettings.valid_directory_name("valid_dir_name") is True

        # Platform-specific path separators
        forward_slash_valid = ProgramSettings.valid_directory_name("valid_dir_name/")
        assert forward_slash_valid is (platform.system() != "Windows")

        backward_slash_valid = ProgramSettings.valid_directory_name("valid_dir_name\\")
        assert backward_slash_valid is (platform.system() == "Windows")

        # Invalid characters should fail validation
        assert ProgramSettings.valid_directory_name("invalid<dir>name") is False
        assert ProgramSettings.valid_directory_name("invalid*dir?name") is False


class TestUserConfigurationAccess:
    """Test user configuration directory access and settings management."""

    def test_user_can_access_existing_config_directory(self, mock_user_config) -> None:
        """
        User can successfully access their configuration directory when it exists.

        GIVEN: A user has a valid configuration directory on their system
        WHEN: User attempts to access their configuration settings
        THEN: The correct configuration directory path should be returned
        """
        # Arrange: Mock existing configuration directory
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.user_config_dir",
                return_value=mock_user_config["config_dir"],
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
            patch("os.path.isdir", return_value=True),
        ):
            # Act: Access user configuration directory
            result = ProgramSettings._user_config_dir()

            # Assert: Correct directory path is returned
            assert result == mock_user_config["config_dir"]

    def test_user_gets_error_when_config_directory_missing(self) -> None:
        """
        User receives appropriate error when configuration directory doesn't exist.

        GIVEN: A user's configuration directory has been deleted or moved
        WHEN: User attempts to access their configuration settings
        THEN: A FileNotFoundError should be raised with helpful information
        """
        # Arrange: Mock missing configuration directory
        with (
            patch("platformdirs.user_config_dir", return_value="/missing/config"),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False),
            pytest.raises(FileNotFoundError),
        ):
            # Act: Attempt to access missing configuration directory
            ProgramSettings._user_config_dir()

    def test_user_gets_error_when_config_path_is_not_directory(self) -> None:
        """
        User receives appropriate error when config path exists but is not a directory.

        GIVEN: A user's configuration path exists but is a file instead of directory
        WHEN: User attempts to access their configuration settings
        THEN: A NotADirectoryError should be raised
        """
        # Arrange: Mock configuration path as file instead of directory
        with (
            patch("platformdirs.user_config_dir", return_value="/path/to/file"),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
            patch("os.path.isdir", return_value=False),
            pytest.raises(NotADirectoryError),
        ):
            # Act: Attempt to access configuration file as directory
            ProgramSettings._user_config_dir()

    def test_user_can_access_site_config_directory(self) -> None:
        """
        User can successfully access the site configuration directory when it exists.

        GIVEN: A user needs to access site-wide configuration templates
        WHEN: User attempts to access the site configuration directory
        THEN: The correct site configuration directory path should be returned
        """
        # Arrange: Mock existing site configuration directory
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.site_config_dir",
                return_value="/site/config/dir",
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
            patch("os.path.isdir", return_value=True),
        ):
            # Act: Access site configuration directory
            result = ProgramSettings._site_config_dir()

            # Assert: Correct directory path is returned
            assert result == "/site/config/dir"

    def test_user_gets_error_when_site_config_directory_missing(self) -> None:
        """
        User receives appropriate error when site configuration directory doesn't exist.

        GIVEN: A user's site configuration directory has been deleted or corrupted
        WHEN: User attempts to access site configuration
        THEN: A FileNotFoundError should be raised with helpful information
        """
        # Arrange: Mock missing site configuration directory
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.site_config_dir",
                return_value="/missing/site/config",
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False),
            pytest.raises(FileNotFoundError, match=r"site configuration directory.*does not exist"),
        ):
            # Act: Attempt to access missing site configuration directory
            ProgramSettings._site_config_dir()

    def test_user_gets_error_when_site_config_path_is_not_directory(self) -> None:
        """
        User receives appropriate error when site config path exists but is not a directory.

        GIVEN: A user's site configuration path exists but is a file instead of directory
        WHEN: User attempts to access site configuration
        THEN: A NotADirectoryError should be raised
        """
        # Arrange: Mock site configuration path as file instead of directory
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.site_config_dir",
                return_value="/path/to/site/config/file",
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
            patch("os.path.isdir", return_value=False),
            pytest.raises(NotADirectoryError, match=r"path.*is not a directory"),
        ):
            # Act: Attempt to access site configuration file as directory
            ProgramSettings._site_config_dir()


class TestSettingsFileOperations:
    """Test user settings file reading and writing operations."""

    def test_user_can_load_existing_settings_file(self, mock_user_config, sample_program_settings) -> None:
        """
        User can successfully load their existing settings from configuration file.

        GIVEN: A user has a valid settings file with saved preferences
        WHEN: User starts the application and settings are loaded
        THEN: All saved settings should be loaded correctly
        AND: Any missing default settings should be added automatically
        """
        # Arrange: Mock existing settings file with realistic data
        expected_result = sample_program_settings.copy()
        # Add all missing defaults that would be added by _recursive_merge_defaults
        expected_result["annotate_docs_into_param_files"] = False  # Added by default
        expected_result["gui_complexity"] = "simple"  # Added by default
        expected_result["motor_test"] = {"duration": 2, "throttle_pct": 10}  # Added by default
        expected_result["display_usage_popup"]["component_editor_validation"] = True  # Added by default

        # Update directory_selection with the defaults that would be merged in
        expected_result["directory_selection"]["new_base_dir"] = os_path.join(mock_user_config["config_dir"], "vehicles")
        expected_result["directory_selection"]["vehicle_dir"] = os_path.join(mock_user_config["config_dir"], "vehicles")

        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value=mock_user_config["config_dir"]),
            patch.object(ProgramSettings, "get_templates_base_dir", return_value="/app/templates"),
            patch("builtins.open", mock_open(read_data=json.dumps(sample_program_settings))),
        ):
            # Act: Load settings from file
            result = ProgramSettings._get_settings_as_dict()

            # Assert: All settings loaded correctly with defaults added
            assert result == expected_result

    def test_user_gets_default_settings_when_file_missing(self, mock_user_config) -> None:
        """
        User receives sensible default settings when configuration file doesn't exist.

        GIVEN: A user starts the application for the first time (no settings file)
        WHEN: Settings are loaded from a non-existent file
        THEN: Complete default configuration should be provided
        AND: All required settings sections should be present
        """
        # Arrange: Mock missing settings file
        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value=mock_user_config["config_dir"]),
            patch("os.path.join", return_value=mock_user_config["settings_file"]),
            patch("builtins.open", side_effect=FileNotFoundError),
        ):
            # Act: Load settings from missing file
            result = ProgramSettings._get_settings_as_dict()

            # Assert: All required default settings are present
            assert "Format version" in result
            assert "directory_selection" in result
            assert "display_usage_popup" in result
            assert "component_editor" in result["display_usage_popup"]
            assert "parameter_editor" in result["display_usage_popup"]
            assert "auto_open_doc_in_browser" in result
            assert "annotate_docs_into_param_files" in result

    def test_user_gets_default_popup_settings_when_missing(self, mock_user_config) -> None:
        """
        User gets default popup settings when display_usage_popup section is incomplete.

        GIVEN: A user has settings file missing specific popup setting keys
        WHEN: User loads their settings
        THEN: Missing popup settings should be initialized with defaults
        AND: Each popup type should get appropriate default values
        """
        # Arrange: Mock settings file missing popup settings details
        incomplete_settings = {
            "Format version": 1,
            "display_usage_popup": {},  # Empty popup settings
        }

        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value=mock_user_config["config_dir"]),
            patch("os.path.join", return_value=mock_user_config["settings_file"]),
            patch("builtins.open", mock_open(read_data=json.dumps(incomplete_settings))),
        ):
            # Act: Load settings with missing popup details
            result = ProgramSettings._get_settings_as_dict()

            # Assert: Popup settings initialized with defaults
            assert result["display_usage_popup"]["component_editor"] is True
            assert result["display_usage_popup"]["parameter_editor"] is True

    def test_user_can_load_settings_from_file_directly(self, mock_user_config) -> None:  # pylint: disable=unused-argument
        """
        User can load settings from file using the file loading method.

        GIVEN: A user has a settings file with specific content
        WHEN: Settings are loaded directly from the file
        THEN: The exact file content should be returned
        """
        # Arrange: Mock settings file content
        file_content = {"test_setting": "test_value"}

        with patch("builtins.open", mock_open(read_data=json.dumps(file_content))):
            # Act: Load settings from file
            result = ProgramSettings._load_settings_from_file("/mock/settings.json")

            # Assert: File content loaded correctly
            assert result == file_content

    def test_user_gets_empty_dict_when_file_missing_during_direct_load(self) -> None:
        """
        User gets empty dictionary when loading settings from missing file.

        GIVEN: A user attempts to load settings from a non-existent file
        WHEN: The direct file loading method is called
        THEN: An empty dictionary should be returned
        """
        # Arrange: Mock missing file
        with patch("builtins.open", side_effect=FileNotFoundError("File not found")):
            # Act: Load settings from missing file
            result = ProgramSettings._load_settings_from_file("/missing/settings.json")

            # Assert: Empty dictionary returned
            assert result == {}

    def test_user_can_apply_default_settings_to_incomplete_settings(self) -> None:
        """
        User can apply default settings to an incomplete settings dictionary.

        GIVEN: A user has a partial settings dictionary missing required keys
        WHEN: Default settings are applied
        THEN: All missing keys should be added with appropriate defaults
        """
        # Arrange: Incomplete settings missing various keys and mock defaults
        incomplete_settings = {"existing_key": "existing_value"}

        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value="/mock/config"),
            patch.object(ProgramSettings, "get_templates_base_dir", return_value="/app/templates"),
        ):
            # Get the defaults from the actual method
            defaults = ProgramSettings._get_settings_defaults()

            # Act: Apply default settings using recursive merge
            result = ProgramSettings._recursive_merge_defaults(incomplete_settings, defaults)

            # Assert: All defaults added while preserving existing
            assert result["existing_key"] == "existing_value"  # Preserved
            assert result["Format version"] == 1  # Added
            assert "directory_selection" in result  # Added
            assert "display_usage_popup" in result  # Added
            assert result["display_usage_popup"]["component_editor"] is True  # Added
            assert result["display_usage_popup"]["parameter_editor"] is True  # Added
            assert result["auto_open_doc_in_browser"] is True  # Added
            assert result["annotate_docs_into_param_files"] is False  # Added

    def test_user_can_normalize_path_separators_for_platform(self) -> None:
        """
        User can normalize path separators to match the current platform.

        GIVEN: A user has file paths with mixed or incorrect separators
        WHEN: Path normalization is applied
        THEN: Separators should be corrected for the current platform
        """
        # Act & Assert: Test path normalization based on platform
        if platform.system() == "Windows":
            # Windows should use backslashes
            assert ProgramSettings._normalize_path_separators("C:/path/to/file") == "C:\\path\\to\\file"
            assert ProgramSettings._normalize_path_separators("C:\\path\\to\\file") == "C:\\path\\to\\file"
        else:
            # Unix-like should use forward slashes
            assert ProgramSettings._normalize_path_separators("C:\\path\\to\\file") == "C:/path/to/file"
            assert ProgramSettings._normalize_path_separators("C:/path/to/file") == "C:/path/to/file"

    def test_user_can_save_settings_to_file(self, mock_user_config) -> None:
        """
        User can successfully save their current settings to configuration file.

        GIVEN: A user has modified their application settings
        WHEN: Settings need to be persisted to disk
        THEN: Settings should be written to the configuration file in JSON format
        AND: File should be created with proper encoding
        """
        # Arrange: Mock settings data and file operations
        mock_settings = {"test_setting": "test_value"}

        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value=mock_user_config["config_dir"]),
            patch("os.path.join", return_value=mock_user_config["settings_file"]),
            patch("builtins.open", mock_open()) as mock_file,
        ):
            # Act: Save settings to file
            ProgramSettings._set_settings_from_dict(mock_settings)

            # Assert: File is opened correctly and data is written
            mock_file.assert_called_once_with(mock_user_config["settings_file"], "w", encoding="utf-8")
            mock_file().write.assert_called()


class TestUsagePopupSettings:
    """Test user interface popup display preferences management."""

    def test_user_can_get_usage_popup_display_preferences(self) -> None:
        """
        User can check their preferences for displaying usage popup windows.

        GIVEN: A user has configured their popup display preferences
        WHEN: User interface needs to decide whether to show popup help
        THEN: Correct preference should be returned for each popup type
        AND: Unknown popup types should default to showing help
        """
        # Arrange: Mock user preferences for popup display
        with patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings:
            mock_get_settings.return_value = {"display_usage_popup": {"component_editor": False, "parameter_editor": True}}

            # Act & Assert: Check specific popup preferences
            assert ProgramSettings.display_usage_popup("component_editor") is False
            assert ProgramSettings.display_usage_popup("parameter_editor") is True
            assert ProgramSettings.display_usage_popup("nonexistent_type") is True  # Default behavior

    def test_user_can_set_usage_popup_display_preferences(self) -> None:
        """
        User can modify their preferences for displaying usage popup windows.

        GIVEN: A user wants to change popup display preferences
        WHEN: User sets a preference for a specific popup type
        THEN: The setting should be saved with the new preference
        AND: Invalid popup types should be ignored for safety
        """
        # Arrange: Mock current settings and save functionality
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
        ):
            mock_get_settings.return_value = {"display_usage_popup": {"component_editor": True, "parameter_editor": True}}

            # Act: Set valid popup preference
            ProgramSettings.set_display_usage_popup("component_editor", value=False)

            # Assert: Settings are updated correctly
            mock_set_settings.assert_called_with(
                {"display_usage_popup": {"component_editor": False, "parameter_editor": True}}
            )

            # Act: Try to set invalid popup type
            mock_set_settings.reset_mock()
            ProgramSettings.set_display_usage_popup("nonexistent_type", value=False)

            # Assert: Invalid types are ignored
            mock_set_settings.assert_not_called()


class TestGenericSettingsAccess:
    """Test access to general settings in ProgramSettings."""

    def test_user_can_read_existing_settings_values(self) -> None:
        """
        User can retrieve values for any stored configuration setting.

        GIVEN: A user has various settings stored in their configuration
        WHEN: User requests specific setting values
        THEN: Correct values should be returned for existing settings
        AND: Default values should be returned for missing settings
        """
        # Arrange: Mock stored settings data
        with patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings:
            mock_get_settings.return_value = {
                "Format version": 2,
                "auto_open_doc_in_browser": False,
                "motor_test_duration": 3.5,
                # Add primitive value directly to test line 304
                "primitive_dict": {},
            }

            # Act & Assert: Read various setting types
            assert ProgramSettings.get_setting("Format version") == 2
            assert ProgramSettings.get_setting("auto_open_doc_in_browser") is False
            assert ProgramSettings.get_setting("motor_test_duration") == 3.5
            assert ProgramSettings.get_setting("nonexistent_setting") is None
            # Test direct primitive value to cover line 304
            assert ProgramSettings.get_setting("primitive_dict") is None

    def test_user_can_write_setting_values(self) -> None:
        """
        User can save new values for any configuration setting.

        GIVEN: A user wants to update a configuration setting
        WHEN: User sets a new value for a valid setting
        THEN: The setting should be saved with the new value
        AND: Invalid settings should be ignored for safety
        """
        # Arrange: Mock settings operations and defaults
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings_dict,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch.object(
                ProgramSettings,
                "_get_settings_defaults",
                return_value={
                    "Format version": 1,
                    "auto_open_doc_in_browser": True,
                    "annotate_docs_into_param_files": False,
                    "gui_complexity": "normal",
                    "motor_test_duration": 2.5,
                    "motor_test_throttle_pct": 10,
                },
            ),
        ):
            mock_get_settings_dict.return_value = {
                "Format version": 1,
                "auto_open_doc_in_browser": True,
                "annotate_docs_into_param_files": False,
                "gui_complexity": "normal",
                "motor_test_duration": 2.5,
                "motor_test_throttle_pct": 10,
            }

            # Act: Set valid setting value
            ProgramSettings.set_setting("gui_complexity", "simple")

            # Assert: Valid setting is updated and saved
            # pylint: disable=duplicate-code
            expected_settings = {
                "Format version": 1,
                "auto_open_doc_in_browser": True,
                "annotate_docs_into_param_files": False,
                "gui_complexity": "simple",
                "motor_test_duration": 2.5,
                "motor_test_throttle_pct": 10,
            }
            # pylint: enable=duplicate-code
            mock_set_settings.assert_called_with(expected_settings)

            # Act: Try to set invalid setting
            mock_set_settings.reset_mock()
            ProgramSettings.set_setting("invalid_setting", "value")

            # Assert: Invalid settings are ignored
            mock_set_settings.assert_not_called()

    def test_user_can_write_nested_setting_values(self) -> None:
        """
        User can update nested settings using hierarchical paths.

        GIVEN: A system with nested settings
        WHEN: User sets a value using a path like "directory_selection/template_dir"
        THEN: The setting should be updated correctly in the nested structure
        """
        # Arrange: Mock settings operations and defaults with nested structure
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings_dict,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
        ):
            mock_get_settings_dict.return_value = {
                "directory_selection": {
                    "template_dir": "/original/path",
                    "new_base_dir": "/base/dir",
                }
            }

            # Act: Set nested setting value
            ProgramSettings.set_setting("directory_selection/template_dir", "/new/path")

            # Assert: Nested setting is updated and saved
            expected_settings = {
                "directory_selection": {
                    "template_dir": "/new/path",
                    "new_base_dir": "/base/dir",
                }
            }
            mock_set_settings.assert_called_with(expected_settings)

    def test_user_gets_error_when_invalid_nested_path(self) -> None:
        """
        User gets error when trying to set a value with invalid nested path.

        GIVEN: A system with settings
        WHEN: User tries to set a value with a nonexistent parent path
        THEN: No setting should be updated and an error should be logged
        """
        # Arrange: Mock settings operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings_dict,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_error") as mock_logging_error,
        ):
            mock_get_settings_dict.return_value = {
                "directory_selection": {
                    "template_dir": "/original/path",
                }
            }

            # Act: Try to set nested setting with invalid parent path
            ProgramSettings.set_setting("nonexistent/template_dir", "/new/path")

            # Assert: Setting is not updated and error is logged
            mock_set_settings.assert_not_called()
            mock_logging_error.assert_called()

    def test_user_gets_error_when_invalid_nested_key(self) -> None:
        """
        User gets error when trying to set a value with valid parent but invalid key.

        GIVEN: A system with nested settings
        WHEN: User tries to set a value with a nonexistent key in a valid parent
        THEN: No setting should be updated and an error should be logged
        """
        # Arrange: Mock settings operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings_dict,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_error") as mock_logging_error,
        ):
            mock_get_settings_dict.return_value = {
                "directory_selection": {
                    "template_dir": "/original/path",
                }
            }

            # Act: Try to set nonexistent nested setting
            ProgramSettings.set_setting("directory_selection/nonexistent_key", "/new/path")

            # Assert: Setting is not updated and error is logged
            mock_set_settings.assert_not_called()
            mock_logging_error.assert_called()

    def test_user_gets_error_when_parent_not_dict(self) -> None:
        """
        User gets error when trying to set a nested path where parent is not a dictionary.

        GIVEN: A system with settings where a primitive value exists
        WHEN: User tries to treat a primitive value as a dictionary
        THEN: No setting should be updated and an error should be logged
        """
        # Arrange: Mock settings operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings_dict,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_error") as mock_logging_error,
        ):
            mock_get_settings_dict.return_value = {
                "primitive_value": "string",
            }

            # Act: Try to use primitive value as dictionary
            ProgramSettings.set_setting("primitive_value/some_key", "value")

            # Assert: Setting is not updated and error is logged
            mock_set_settings.assert_not_called()
            mock_logging_error.assert_called()


class TestMotorDiagramManagement:
    """Test motor diagram file management functionality in ProgramSettings."""

    def test_user_can_get_motor_diagram_filepath_for_existing_frame(self) -> None:
        """
        User can get the correct motor diagram filepath for existing frame configuration.

        GIVEN: A vehicle with quad X frame configuration (class=1, type=1)
        WHEN: User requests the motor diagram filepath
        THEN: The correct SVG filepath should be returned
        AND: The file should exist in the images directory
        """
        # Arrange: Mock file operations to simulate existing diagram
        mock_svg_path = "/app/images/m_01_01_quad_x.svg"

        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.glob_glob", return_value=[mock_svg_path]
            ),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
        ):
            # Act: Get diagram filepath for quad X frame
            filepath, error_msg = ProgramSettings.motor_diagram_filepath(frame_class=1, frame_type=1)

            # Assert: Correct filepath is returned without error
            assert filepath == mock_svg_path
            assert error_msg == ""
            assert "m_01_01" in filepath
            assert filepath.endswith(".svg")

    def test_user_gets_default_diagram_when_specific_frame_not_found(self) -> None:
        """
        User gets default quad X diagram when specific frame diagram doesn't exist.

        GIVEN: A vehicle with an uncommon frame configuration
        WHEN: User requests motor diagram filepath for non-existent frame
        THEN: The default quad X diagram should be returned as fallback
        """
        # Arrange: Mock file operations with no specific match but default exists
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.glob_glob", return_value=[]
            ),  # No specific match
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True
            ),  # Default exists
        ):
            # Act: Get diagram filepath for uncommon frame
            filepath, error_msg = ProgramSettings.motor_diagram_filepath(frame_class=99, frame_type=99)

            # Assert: Error message indicates diagram not found
            assert filepath == ""
            assert error_msg != ""
            assert "not found" in error_msg

    def test_user_gets_first_file_when_multiple_diagrams_found(self) -> None:
        """
        User gets the first matching file when multiple diagram files exist.

        GIVEN: A system with multiple matching motor diagram files
        WHEN: User requests motor diagram filepath for a frame with multiple matches
        THEN: The first matching file should be returned with an error message
        """
        # Arrange: Mock file operations with multiple matches
        multiple_matches = ["/app/images/m_01_01_quad_x_v1.svg", "/app/images/m_01_01_quad_x_v2.svg"]

        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.glob_glob", return_value=multiple_matches
            ),  # Multiple matches
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_error") as mock_logging_error,
        ):
            # Act: Get diagram filepath
            filepath, error_msg = ProgramSettings.motor_diagram_filepath(frame_class=1, frame_type=1)

            # Assert: First match returned with error message
            assert filepath == multiple_matches[0]
            assert "Multiple motor diagrams found" in error_msg
            # Verify that logging_error was called (line 355)
            mock_logging_error.assert_called_once()

    def test_user_gets_empty_string_when_no_diagrams_exist(self) -> None:
        """
        User gets empty string when no motor diagrams exist at all.

        GIVEN: A system with no motor diagram files
        WHEN: User requests motor diagram filepath
        THEN: An empty string should be returned
        """
        # Arrange: Mock file operations with no diagrams existing
        with (
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.glob_glob", return_value=[]
            ),  # No specific match
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False
            ),  # No default either
        ):
            # Act: Get diagram filepath when no diagrams exist
            filepath, error_msg = ProgramSettings.motor_diagram_filepath(frame_class=1, frame_type=1)

            # Assert: Empty string is returned with error message
            assert filepath == ""
            assert error_msg != ""
            assert "not found" in error_msg

    def test_user_can_check_motor_diagram_existence(self) -> None:
        """
        User can check if motor diagram exists for specific frame configuration.

        GIVEN: A user wants to verify motor diagram availability
        WHEN: User checks if diagram exists for specific frame
        THEN: Boolean result should indicate diagram availability correctly
        """
        # Arrange: Mock diagram filepath and existence check
        with patch.object(ProgramSettings, "motor_diagram_filepath") as mock_filepath:
            # Act & Assert: Test when diagram exists
            # motor_diagram_filepath still returns a tuple (path, error_msg) as per its implementation
            mock_filepath.return_value = "/app/images/m_01_01_quad_x.svg", ""
            with patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True
            ):
                assert ProgramSettings.motor_diagram_exists(1, 1) is True

            # Act & Assert: Test when diagram doesn't exist
            mock_filepath.return_value = "", "Not found"
            assert ProgramSettings.motor_diagram_exists(99, 99) is False

    def test_motor_diagram_filepath_works_in_compiled_executable(self) -> None:
        """
        Motor diagram filepath works correctly when running as compiled executable.

        GIVEN: Application running as compiled executable (frozen)
        WHEN: User requests motor diagram filepath
        THEN: Correct path relative to executable should be returned
        """
        # Arrange: Mock frozen application environment
        mock_svg_path = "/compiled/app/images/m_01_01_quad_x.svg"

        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.getattr") as mock_getattr,
            patch("sys.executable", "/compiled/app/ardupilot_methodic_configurator.exe"),
            patch("os.path.dirname", return_value="/compiled/app"),
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.glob_glob", return_value=[mock_svg_path]
            ),
        ):
            # Configure mock to return True for frozen attribute, False for others
            def getattr_side_effect(_obj: object, attr: str, default: object = None) -> object:
                if attr == "frozen":
                    return True
                return default if default is not None else False

            mock_getattr.side_effect = getattr_side_effect
            # Act: Get diagram filepath in compiled environment
            filepath, error_msg = ProgramSettings.motor_diagram_filepath(frame_class=1, frame_type=1)

            # Assert: Correct compiled app path is used
            assert filepath == mock_svg_path
            assert error_msg == ""


class TestTemplateDirectoryManagement:
    """Test template directory storage and retrieval functionality."""

    def test_user_can_get_recently_used_dirs_with_directory_creation(self) -> None:
        """
        User can retrieve recently used directories and system creates missing directories.

        GIVEN: A user requests their recently used directory paths
        WHEN: The default vehicles directory does not exist
        THEN: The vehicles directory should be created automatically
        AND: All three directory paths should be returned correctly
        """
        # Arrange: Mock missing vehicles directory that needs creation
        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value="/user/config"),
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=False),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_makedirs") as mock_makedirs,
        ):
            # Configure settings response
            mock_get_settings.return_value = {
                "directory_selection": {
                    "template_dir": "/saved/template",
                    "new_base_dir": "/saved/base",
                    "vehicle_dir": "/saved/vehicle",
                }
            }

            # Act: Get recently used directories
            template_dir, new_base_dir, vehicle_dir = ProgramSettings.get_recently_used_dirs()

            # Assert: Directory was created and paths returned correctly
            # Use os_path.join to create platform-appropriate path for assertion
            expected_path = os_path.join("/user/config", "vehicles")
            mock_makedirs.assert_called_once_with(expected_path, exist_ok=True)
            assert template_dir == "/saved/template"
            assert new_base_dir == "/saved/base"
            assert vehicle_dir == "/saved/vehicle"

    def test_user_can_get_recently_used_dirs_without_directory_creation(self) -> None:
        """
        User can retrieve recently used directories when vehicles directory already exists.

        GIVEN: A user requests their recently used directory paths
        WHEN: The default vehicles directory already exists
        THEN: No directory creation occurs and paths are returned correctly
        """
        # Arrange: Mock existing vehicles directory (no creation needed)
        with (
            patch.object(ProgramSettings, "_user_config_dir", return_value="/user/config"),
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path.exists", return_value=True),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_makedirs") as mock_makedirs,
        ):
            # Configure settings response
            mock_get_settings.return_value = {
                "directory_selection": {
                    "template_dir": "/saved/template",
                    "new_base_dir": "/saved/base",
                    "vehicle_dir": "/saved/vehicle",
                }
            }

            # Act: Get recently used directories
            template_dir, new_base_dir, vehicle_dir = ProgramSettings.get_recently_used_dirs()

            # Assert: No directory creation occurred
            mock_makedirs.assert_not_called()
            assert template_dir == "/saved/template"
            assert new_base_dir == "/saved/base"
            assert vehicle_dir == "/saved/vehicle"

    def test_user_can_store_template_directory(self) -> None:
        """
        User can store their preferred template directory for vehicle creation.

        GIVEN: A user selects a template directory for their projects
        WHEN: User stores the template directory path
        THEN: The directory path should be saved correctly with proper formatting
        """
        # Arrange: Mock setting operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch.object(ProgramSettings, "get_templates_base_dir", return_value="/app/templates"),
            patch("os.path.join", return_value="/app/templates/Copter/QuadX"),
        ):
            mock_get_settings.return_value = {"directory_selection": {"template_dir": "/old/template"}}

            # Act: Store new template directory
            ProgramSettings.store_template_dir("Copter/QuadX")

            # Assert: Setting was called with correct structure
            mock_set_settings.assert_called_once()
            # Get the called arguments
            called_args = mock_set_settings.call_args[0][0]
            assert "directory_selection" in called_args
            # Path will be normalized for current platform
            expected_path = "/app/templates/Copter/QuadX"
            if platform.system() == "Windows":
                expected_path = "\\app\\templates\\Copter\\QuadX"
            assert called_args["directory_selection"]["template_dir"] == expected_path

    def test_user_can_store_recently_used_template_dirs(self) -> None:
        """
        User can store recently used template and base directories.

        GIVEN: A user has selected template and base directories for their projects
        WHEN: User stores these directory paths
        THEN: Both paths should be saved with proper normalization
        """
        # Arrange: Mock setting operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
        ):
            mock_get_settings.return_value = {"directory_selection": {}}

            # Act: Store recently used directories
            ProgramSettings.store_recently_used_template_dirs("/template/path", "/base/path")

            # Assert: Both paths saved with normalization
            mock_set_settings.assert_called_once()
            called_args = mock_set_settings.call_args[0][0]
            assert "template_dir" in called_args["directory_selection"]
            assert "new_base_dir" in called_args["directory_selection"]

    def test_user_can_store_recently_used_vehicle_dir(self) -> None:
        """
        User can store their recently used vehicle directory.

        GIVEN: A user has selected a vehicle directory for their projects
        WHEN: User stores the vehicle directory path
        THEN: The path should be saved with proper normalization
        """
        # Arrange: Mock setting operations
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
        ):
            mock_get_settings.return_value = {"directory_selection": {}}

            # Act: Store recently used vehicle directory
            ProgramSettings.store_recently_used_vehicle_dir("/vehicle/path")

            # Assert: Vehicle path saved with normalization
            mock_set_settings.assert_called_once()
            called_args = mock_set_settings.call_args[0][0]
            assert "vehicle_dir" in called_args["directory_selection"]


class TestGUIComplexitySettings:
    """Test GUI complexity preference management."""

    def test_user_can_manage_gui_complexity_setting(self) -> None:
        """
        User can control the GUI complexity level for interface simplification.

        GIVEN: A user wants to configure interface complexity
        WHEN: User gets and sets the GUI complexity preference
        THEN: The preference should be stored and retrieved correctly
        AND: Default value should be provided when setting doesn't exist
        """
        # Arrange: Mock setting operations and defaults
        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get_settings,
            patch.object(ProgramSettings, "_set_settings_from_dict") as mock_set_settings,
            patch.object(
                ProgramSettings,
                "_get_settings_defaults",
                return_value={
                    "Format version": 1,
                    "auto_open_doc_in_browser": True,
                    "annotate_docs_into_param_files": False,
                    "gui_complexity": "normal",
                    "motor_test_duration": 2.5,
                    "motor_test_throttle_pct": 10,
                },
            ),
        ):
            # Test getting existing setting
            mock_get_settings.return_value = {"gui_complexity": "simple"}
            complexity = ProgramSettings.get_setting("gui_complexity")
            assert complexity == "simple"

            # Test getting default when setting doesn't exist
            mock_get_settings.return_value = {}
            complexity = ProgramSettings.get_setting("gui_complexity")
            assert complexity is None  # Returns None for missing settings

            # Test setting GUI complexity
            mock_get_settings.return_value = {
                "Format version": 1,
                "auto_open_doc_in_browser": True,
                "annotate_docs_into_param_files": False,
                "gui_complexity": "normal",
                "motor_test_duration": 2.5,
                "motor_test_throttle_pct": 10,
            }
            ProgramSettings.set_setting("gui_complexity", "simple")
            # pylint: disable=duplicate-code
            expected_settings = {
                "Format version": 1,
                "auto_open_doc_in_browser": True,
                "annotate_docs_into_param_files": False,
                "gui_complexity": "simple",
                "motor_test_duration": 2.5,
                "motor_test_throttle_pct": 10,
            }
            # pylint: disable=duplicate-code
            mock_set_settings.assert_called_with(expected_settings)


class TestInternalConfigurationMethods:
    """Test internal configuration helper methods for complete coverage."""

    def test_get_templates_base_dir_uses_script_dir_on_linux(self) -> None:
        """
        Templates base directory uses package path on Linux platform.

        GIVEN: A Linux platform environment
        WHEN: The templates base directory is requested
        THEN: The package path should be used as the base
        """
        # Arrange: Mock platform and importlib_files
        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.importlib_files") as mock_importlib,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_debug"),
        ):
            # Test for Linux platform
            mock_platform.return_value = "Linux"
            mock_package = MagicMock()
            mock_importlib.return_value = mock_package
            mock_package.__truediv__ = MagicMock()
            mock_package.__truediv__.return_value.__str__ = MagicMock(return_value="/mock/package/path/vehicle_templates")

            # Act: Get templates base directory
            result = ProgramSettings.get_templates_base_dir()

            # Assert: Package path is used correctly
            assert result == "/mock/package/path/vehicle_templates"
            mock_importlib.assert_called_once_with("ardupilot_methodic_configurator")

    def test_get_templates_base_dir_uses_script_dir_on_macos(self) -> None:
        """
        Templates base directory uses package path on macOS platform.

        GIVEN: A macOS platform environment
        WHEN: The templates base directory is requested
        THEN: The package path should be used as the base
        """
        # Arrange: Mock macOS platform and importlib_files
        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.importlib_files") as mock_importlib,
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.logging_debug"),
        ):
            # Test for macOS platform (Darwin)
            mock_platform.return_value = "Darwin"
            mock_package = MagicMock()
            mock_importlib.return_value = mock_package
            mock_package.__truediv__ = MagicMock()
            mock_package.__truediv__.return_value.__str__ = MagicMock(return_value="/mock/package/path/vehicle_templates")

            # Act: Get templates base directory
            result = ProgramSettings.get_templates_base_dir()

            # Assert: Package path is used correctly (same as Linux - non-Windows)
            assert result == "/mock/package/path/vehicle_templates"
            mock_importlib.assert_called_once_with("ardupilot_methodic_configurator")

    def test_get_templates_base_dir_uses_site_config_on_windows(self) -> None:
        """
        Templates base directory uses site config directory on Windows.

        GIVEN: A Windows platform environment
        WHEN: The templates base directory is requested
        THEN: The site configuration directory should be used as the base
        """
        # Arrange: Mock Windows platform and site config
        with (
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform,
            patch(
                "ardupilot_methodic_configurator.backend_filesystem_program_settings.ProgramSettings._site_config_dir"
            ) as mock_site_config,
        ):
            mock_platform.return_value = "Windows"
            mock_site_config.return_value = "C:\\Program Files\\App"

            # Act: Get templates base directory
            result = ProgramSettings.get_templates_base_dir()

            # Assert: Windows path logic is used correctly
            mock_site_config.assert_called_once()
            # Use os.path.join to handle platform-specific path separators
            expected_path = os_path.join("C:\\Program Files\\App", "vehicle_templates")
            assert result == expected_path



"""
TKinter base classes reused in multiple parts of the code.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

# https://wiki.tcl-lang.org/page/Changing+Widget+Colors

import tkinter as tk
from platform import system as platform_system
from tkinter import messagebox, ttk
from typing import Optional

from ardupilot_methodic_configurator import _


def show_error_message(title: str, message: str) -> None:
    root = tk.Tk(className="ArduPilotMethodicConfigurator")
    # Set the theme to 'alt'
    style = ttk.Style()
    style.theme_use("alt")
    root.withdraw()  # Hide the main window
    messagebox.showerror(title, message)
    root.destroy()


def show_warning_message(title: str, message: str) -> None:
    root = tk.Tk(className="ArduPilotMethodicConfigurator")
    # Set the theme to 'alt'
    style = ttk.Style()
    style.theme_use("alt")
    root.withdraw()  # Hide the main window
    messagebox.showwarning(title, message)
    root.destroy()


def show_no_param_files_error(_dirname: str) -> None:
    error_message = _(
        "No intermediate parameter files found in the selected '{_dirname}' vehicle directory.\n"
        "Please select and step inside a vehicle directory containing valid ArduPilot intermediate parameter files.\n\n"
        "Make sure to step inside the directory (double-click) and not just select it."
    )
    show_error_message(_("No Parameter Files Found"), error_message.format(**locals()))


def show_no_connection_error(_error_string: str) -> None:
    error_message = _("{_error_string}\n\nPlease connect a flight controller to the PC,\nwait at least 7 seconds and retry.")
    show_error_message(_("No Connection to the Flight Controller"), error_message.format(**locals()))


class Tooltip:
    """
    A tooltip class for displaying tooltips on widgets.

    Creates a tooltip that appears when the mouse hovers over a widget and disappears when the mouse leaves the widget.
    """

    def __init__(self, widget: tk.Widget, text: str, position_below: bool = True) -> None:
        self.widget: tk.Widget = widget
        self.text: str = text
        self.tooltip: Optional[tk.Toplevel] = None
        self.position_below: bool = position_below

        # Bind the <Enter> and <Leave> events to show and hide the tooltip
        if platform_system() == "Darwin":
            # On macOS, only create the tooltip when the mouse enters the widget
            self.widget.bind("<Enter>", self.create_show)
            self.widget.bind("<Leave>", self.destroy_hide)
        else:
            self.widget.bind("<Enter>", self.show)
            self.widget.bind("<Leave>", self.hide)
            # On non-macOS, create the tooltip immediately and show/hide it on events
            self.tooltip = tk.Toplevel(widget)
            self.tooltip.wm_overrideredirect(boolean=True)
            tooltip_label = ttk.Label(
                self.tooltip, text=text, background="#ffffe0", relief="solid", borderwidth=1, justify=tk.LEFT
            )
            tooltip_label.pack()
            self.tooltip.withdraw()  # Initially hide the tooltip

    def show(self, event: Optional[tk.Event] = None) -> None:  # noqa: ARG002 # pylint: disable=unused-argument
        """On non-macOS, tooltip already exists, show it on events."""
        if self.tooltip:
            self.position_tooltip()
            self.tooltip.deiconify()

    def create_show(self, event: Optional[tk.Event] = None) -> None:  # noqa: ARG002 # pylint: disable=unused-argument
        """On macOS, only create the tooltip when the mouse enters the widget."""
        if self.tooltip:
            return  # Avoid redundant tooltip creation

        self.tooltip = tk.Toplevel(self.widget)

        try:
            self.tooltip.tk.call(
                "::tk::unsupported::MacWindowStyle",
                "style",
                self.tooltip._w,  # type: ignore[attr-defined] # noqa: SLF001 # pylint: disable=protected-access
                "help",
                "noActivates",
            )
            self.tooltip.configure(bg="#ffffe0")
        except AttributeError:  # Catches protected member access error
            self.tooltip.wm_attributes("-alpha", 1.0)  # Ensure opacity
            self.tooltip.wm_attributes("-topmost", True)  # Keep on top # noqa: FBT003
            self.tooltip.configure(bg="#ffffe0")
        tooltip_label = ttk.Label(
            self.tooltip, text=self.text, background="#ffffe0", relief="solid", borderwidth=1, justify=tk.LEFT
        )
        tooltip_label.pack()
        self.position_tooltip()

    def position_tooltip(self) -> None:
        """Calculate the position of the tooltip based on the widget's position."""
        x = self.widget.winfo_rootx() + min(self.widget.winfo_width() // 2, 100)
        y = self.widget.winfo_rooty() + (self.widget.winfo_height() if self.position_below else -10)
        if self.tooltip:
            self.tooltip.geometry(f"+{x}+{y}")

    def hide(self, event: Optional[tk.Event] = None) -> None:  # noqa: ARG002 # pylint: disable=unused-argument
        """On non-macOS, tooltip already exists, hide it on events."""
        if self.tooltip:
            self.tooltip.withdraw()

    def destroy_hide(self, event: Optional[tk.Event] = None) -> None:  # noqa: ARG002 # pylint: disable=unused-argument
        """On macOS, fully destroy the tooltip when the mouse leaves the widget."""
        if self.tooltip:
            self.tooltip.destroy()
            self.tooltip = None


def show_tooltip(widget: tk.Widget, text: str, position_below: bool = True) -> Tooltip:
    return Tooltip(widget, text, position_below=position_below)
